# -*- coding: utf8 -*-
"""
.. module:: burpui.api.restore
    :platform: Unix
    :synopsis: Burp-UI restore api module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import select
import struct

from . import api
from ..engines.server import BUIServer  # noqa
from .custom import fields, Resource
from ..exceptions import BUIserverException

from zlib import adler32
from time import gmtime, strftime, time
from flask import Response, send_file, make_response, after_this_request, current_app
from flask_login import current_user
from flask_restx import inputs
from werkzeug.datastructures import Headers
from werkzeug.exceptions import HTTPException

bui = current_app  # type: BUIServer
ns = api.namespace("restore", "Restore methods")


@ns.route(
    "/archive/<name>/<int:backup>",
    "/<server>/archive/<name>/<int:backup>",
    endpoint="restore",
)
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
        "name": "Client name",
        "backup": "Backup number",
    },
)
class Restore(Resource):
    """The :class:`burpui.api.restore.Restore` resource allows you to
    perform a file restoration.

    This resource is part of the :mod:`burpui.api.restore` module.

    The following parameters are supported:
    - ``list``: list of files/directories to restore
    - ``strip``: number of elements to strip in the path
    - ``format``: returning archive format
    - ``pass``: password to use for encrypted backups
    """

    parser = ns.parser()
    parser.add_argument(
        "pass", help="Password to use for encrypted backups", nullable=True
    )
    parser.add_argument(
        "format",
        required=False,
        help="Returning archive format",
        choices=("zip", "tar.gz", "tar.bz2"),
        default="zip",
        nullable=True,
    )
    parser.add_argument(
        "strip",
        type=int,
        help="Number of elements to strip in the path",
        default=0,
        nullable=True,
    )
    parser.add_argument(
        "list",
        required=True,
        help="List of files/directories to restore",
        nullable=False,
    )
    # FIXME: the example json seems interpreted during the raise of the exception
    # parser.add_argument('list', required=True, help='List of files/directories to restore (example: \'{"restore":[{"folder":true,"key":"/etc"}]}\')', nullable=False)

    @ns.expect(parser, validate=True)
    @ns.doc(
        responses={
            200: "Success",
            400: "Missing parameter",
            403: "Insufficient permissions",
            500: "Internal failure",
        },
    )
    def post(self, server=None, name=None, backup=None):
        """Performs an online restoration

        **POST** method provided by the webservice.
        This method returns a :mod:`flask.Response` object.

        :param server: Which server to collect data from when in multi-agent mode
        :type server: str

        :param name: The client we are working on
        :type name: str

        :param backup: The backup we are working on
        :type backup: int

        :returns: A :mod:`flask.Response` object representing an archive of the restored files
        """
        args = self.parser.parse_args()
        lst = args["list"]
        stp = args["strip"]
        fmt = args["format"] or "zip"
        pwd = args["pass"]
        args_log = args.copy()
        # don't leak secrets in logs
        del args_log["pass"]
        bui.audit.logger.info(
            f"requested restoration of backup n°{backup} for {name} with {args_log}",
            server=server,
        )
        resp = None
        # Check params
        if not lst or not name or not backup:
            self.abort(400, "missing arguments")
        # Manage ACL
        if (
            not current_user.is_anonymous
            and not current_user.acl.is_admin()
            and not current_user.acl.is_client_rw(name, server)
        ):
            self.abort(
                403, "You are not allowed to perform a restoration for this client"
            )
        if server:
            filename = "restoration_%d_%s_on_%s_at_%s.%s" % (
                backup,
                name,
                server,
                strftime("%Y-%m-%d_%H_%M_%S", gmtime()),
                fmt,
            )
        else:
            filename = "restoration_%d_%s_at_%s.%s" % (
                backup,
                name,
                strftime("%Y-%m-%d_%H_%M_%S", gmtime()),
                fmt,
            )

        archive, err = bui.client.restore_files(
            name, backup, lst, stp, fmt, pwd, server
        )
        if not archive:
            bui.audit.logger.error(f"restoration failed: {err}")
            if err:
                if (
                    not current_user.is_anonymous
                    and not current_user.acl.is_admin()
                    or bui.demo
                ) and err != "encrypted":
                    err = (
                        "An error occurred while performing the "
                        "restoration. Please contact your administrator "
                        "for further details"
                    )
                return make_response(err, 500)
            return make_response(err, 500)

        if not server:
            try:
                # Trick to delete the file while sending it to the client.
                # First, we open the file in reading mode so that a file handler
                # is open on the file. Then we delete it as soon as the request
                # ended. Because the fh is open, the file will be actually removed
                # when the transfer is done and the send_file method has closed
                # the fh. Only tested on Linux systems.
                fh = open(archive, "rb")

                @after_this_request
                def remove_file(response):
                    """Callback function to run after the client has handled
                    the request to remove temporary files.
                    """
                    import os

                    os.remove(archive)
                    return response

                resp = send_file(
                    fh,
                    as_attachment=True,
                    attachment_filename=filename,
                    mimetype="application/zip",
                )
                resp.set_cookie("fileDownload", "true")
            except Exception as exp:
                bui.client.logger.error(str(exp))
                self.abort(500, str(exp))
        else:
            # Multi-agent mode
            try:
                socket = bui.client.get_file(archive, server)
                if not socket:
                    self.abort(500)

                lengthbuf = socket.recv(8)
                (length,) = struct.unpack("!Q", lengthbuf)

                bui.client.logger.debug(
                    "Need to get {} Bytes : {}".format(length, socket)
                )

                def stream_file(sock, size):
                    """The restoration took place on another server so we need
                    to stream the file that is not present on the current
                    machine.
                    """
                    bsize = 1024
                    received = 0
                    if size < bsize:
                        bsize = size
                    while received < size:
                        buf = b""
                        read, _, _ = select.select([sock], [], [], 5)
                        if not read:
                            raise Exception("Socket timed-out")
                        buf += sock.recv(bsize)
                        if not buf:
                            continue
                        received += len(buf)
                        self.logger.debug("{}/{}".format(received, size))
                        yield buf
                    sock.sendall(struct.pack("!Q", 2))
                    sock.sendall(b"RE")
                    sock.close()

                headers = Headers()
                headers.add("Content-Disposition", "attachment", filename=filename)
                headers["Content-Length"] = length

                resp = Response(
                    stream_file(socket, length),
                    mimetype="application/zip",
                    headers=headers,
                    direct_passthrough=True,
                )
                resp.set_cookie("fileDownload", "true")
                resp.set_etag(
                    "flask-%s-%s-%s"
                    % (time(), length, adler32(filename.encode("utf-8")) & 0xFFFFFFFF)
                )
            except HTTPException as exp:
                raise exp
            except Exception as exp:
                bui.client.logger.error(str(exp))
                self.abort(500, str(exp))
        bui.audit.logger.info(f"sending file {archive}")
        return resp


@ns.route(
    "/server-restore/<name>",
    "/<server>/server-restore/<name>",
    methods=["GET", "DELETE"],
    endpoint="is_server_restore",
)
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
        "name": "Client name",
    },
)
class ServerRestore(Resource):
    """The :class:`burpui.api.restore.ServerRestore` resource allows you to
    monitor or cancel a server-initiated restoration.

    This resource is part of the :mod:`burpui.api.restore` module.
    """

    list_fields = ns.model(
        "ListRestoreFiles",
        {
            "key": fields.String(
                required=True, description="Path to a file/directory to restore"
            ),
            "folder": fields.Boolean(
                required=True, description="Is the path pointed to a directory"
            ),
        },
    )
    restoration_fields = ns.model(
        "EditRestoration",
        {
            "backup": fields.Integer(
                required=True, description="Backup number to restore"
            ),
            "strip": fields.Integer(
                required=False,
                description="Number of leading path to strip while restoring",
            ),
            "prefix": fields.String(
                required=False,
                description="Where to restore files",
                attribute="restoreprefix",
            ),
            "force": fields.Boolean(
                required=False,
                description="Whether to replace files that are already present",
                default=False,
                attribute="overwrite",
            ),
            "to": fields.String(
                required=False,
                description="What client the restoration is intended to",
            ),
            "orig_client": fields.String(
                required=False,
                description="Name of the client to restore from when different"
                + " from the current one",
            ),
            "found": fields.Boolean(
                required=False, description="Did we found a restore file", default=False
            ),
            "list": fields.Nested(
                list_fields, as_list=True, description="List of path to restore"
            ),
        },
    )

    @ns.marshal_with(restoration_fields, code=200, description="Success")
    @ns.doc(
        responses={
            400: "Missing parameter",
            403: "Insufficient permissions",
            500: "Internal failure",
        },
    )
    def get(self, server=None, name=None):
        """Reads the content of the 'restore' file if present

        **GET** method provided by the webservice.

        :param server: Which server to collect data from when in multi-agent
                       mode
        :type server: str

        :param name: The client we are working on
        :type name: str

        :returns: The content of the restore file
        """
        if not name:
            self.abort(400, "Missing options")
        # Manage ACL
        if (
            not current_user.is_anonymous
            and not current_user.acl.is_admin()
            and not current_user.acl.is_client_rw(name, server)
        ):
            self.abort(403, "You are not allowed to edit a restoration for this client")
        try:
            return bui.client.is_server_restore(name, server)
        except BUIserverException as e:
            self.abort(500, str(e))

    @ns.doc(
        responses={
            200: "Success",
            400: "Missing parameter",
            403: "Insufficient permissions",
            500: "Internal failure",
        },
    )
    def delete(self, server=None, name=None):
        """Remove the 'restore' file if present

        **DELETE** method provided by the webservice.

        :param server: Which server to collect data from when in multi-agent
                       mode
        :type server: str

        :param name: The client we are working on
        :type name: str

        :returns: Status message (success or failure)
        """
        bui.audit.logger.info(
            f"attempt to remove 'restore' file for {name}", server=server
        )
        if not name:
            self.abort(400, "Missing options")
        # Manage ACL
        if (
            not current_user.is_anonymous
            and not current_user.acl.is_admin()
            and not current_user.acl.is_client_rw(name, server)
        ):
            self.abort(
                403, "You are not allowed to cancel a restoration for this client"
            )
        try:
            return bui.client.cancel_server_restore(name, server)
        except BUIserverException as e:
            self.abort(500, str(e))


@ns.route(
    "/server-restore/<name>/<int:backup>",
    "/<server>/server-restore/<name>/<int:backup>",
    methods=["PUT"],
    endpoint="server_restore",
)
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
        "name": "Client name",
        "backup": "Backup number",
    },
)
class DoServerRestore(Resource):
    """The :class:`burpui.api.restore.DoServerRestore` resource allows you to
    schedule a server-initiated restoration.

    The following parameters are supported:
    - ``list-sc``: list of files/directories to restore
    - ``strip-sc``: number of elements to strip in the path
    - ``prefix-sc``: prefix to the restore path
    - ``force-sc``: whether to overwrite existing files
    - ``restoreto-sc``: restore files on an other client
    """

    parser = ns.parser()
    parser.add_argument(
        "list-sc",
        required=True,
        help="List of files/directories to restore",
        nullable=False,
    )
    parser.add_argument(
        "strip-sc",
        type=int,
        help="Number of elements to strip in the path",
        default=0,
        nullable=True,
    )
    parser.add_argument("prefix-sc", help="Prefix to the restore path", nullable=True)
    parser.add_argument(
        "force-sc",
        type=inputs.boolean,
        help="Whether to overwrite existing files",
        default=False,
        nullable=True,
    )
    parser.add_argument(
        "restoreto-sc", help="Restore files on an other client", nullable=True
    )

    @api.disabled_on_demo()
    @ns.expect(parser)
    @ns.doc(
        responses={
            201: "Success",
            400: "Missing parameter",
            403: "Insufficient permissions",
            428: "Configuration forbid this request",
            500: "Internal failure",
        },
    )
    def put(self, server=None, name=None, backup=None):
        """Schedule a server-initiated restoration

        **PUT** method provided by the webservice.

        :param server: Which server to collect data from when in multi-agent
                       mode
        :type server: str

        :param name: The client we are working on
        :type name: str

        :param backup: The backup we are working on
        :type backup: int

        :returns: Status message (success or failure)
        """
        args = self.parser.parse_args()
        files_list = args["list-sc"]
        strip = args["strip-sc"]
        prefix = args["prefix-sc"]
        force = args["force-sc"]
        to = args["restoreto-sc"] or name
        json = []

        if (
            not bui.client.get_parser(agent=server).param(
                "server_can_restore", "client_conf"
            )
            and bui.noserverrestore
        ):
            self.abort(
                428,
                "Sorry this method is not available with the current " "configuration",
            )

        # Check params
        if not files_list or not name or not backup:
            self.abort(400, "Missing options")
        # Manage ACL
        if (
            not current_user.is_anonymous
            and not current_user.acl.is_admin()
            and not current_user.acl.is_client_rw(to, server)
            and not current_user.acl.is_client_allowed(to, server)
        ):
            self.abort(
                403, "You are not allowed to perform a restoration for this client"
            )
        try:
            if to == name:
                to = None
            json = bui.client.server_restore(
                name, backup, files_list, strip, force, prefix, to, server
            )
            bui.audit.logger.info(
                f"requested server-initiated restoration from {name} to {to}",
                server=server,
            )
            return json, 201
        except BUIserverException as e:
            self.abort(500, str(e))



# -*- coding: utf8 -*-
"""
.. module:: burpui.api.settings
    :platform: Unix
    :synopsis: Burp-UI settings api module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import json

from . import api
from ..engines.server import BUIServer  # noqa
from ..ext.cache import cache
from .custom import Resource
from .._compat import unquote
from ..utils import NOTIF_INFO

from flask_babel import gettext as _, refresh
from flask import (
    jsonify,
    request,
    url_for,
    current_app,
    g,
    session,
    render_template_string,
)
from flask_login import current_user
from flask_restx import inputs
from jinja2 import Environment, meta
from ..datastructures import ImmutableMultiDict, MultiDict

TEMPLATE_EXCLUDES = ["client", "agent"]

bui = current_app  # type: BUIServer
ns = api.namespace("settings", "Settings methods")


@ns.route(
    "/server-config",
    "/<server>/server-config",
    "/server-config/<path:conf>",
    "/<server>/server-config/<path:conf>",
    endpoint="server_settings",
)
@ns.doc(
    params={
        "conf": "Path of the configuration file",
        "server": "Which server to collect data from when in multi-agent mode",
    },
)
class ServerSettings(Resource):
    """The :class:`burpui.api.settings.ServerSettings` resource allows you to
    read and write the server's configuration.

    This resource is part of the :mod:`burpui.api.settings` module.
    """

    @api.disabled_on_demo()
    @api.acl_admin_required(
        message="Sorry, you don't have rights to access the setting panel"
    )
    @ns.doc(
        responses={
            200: "Success",
            403: "Insufficient permissions",
            500: "Internal failure",
        }
    )
    def post(self, conf=None, server=None):
        """Saves the server configuration"""
        noti = bui.client.store_conf_srv(request.form, conf, server)
        bui.audit.logger.info(
            f"updated burp-server configuration ({conf})", server=server
        )
        return {"notif": noti}, 200

    @api.disabled_on_demo()
    @api.acl_admin_required(
        message="Sorry, you don't have rights to access the setting panel"
    )
    @ns.doc(
        responses={
            200: "Success",
            403: "Insufficient permissions",
            500: "Internal failure",
        }
    )
    def delete(self, conf=None, server=None):
        """Deletes a configuration file"""
        try:
            conf = unquote(conf)
        except:
            pass
        parser = bui.client.get_parser(agent=server)
        bui.audit.logger.info(f"requested removal of {conf}", server=server)
        return parser.remove_conf(conf)

    @api.acl_admin_required(
        message="Sorry, you don't have rights to access the setting panel"
    )
    @ns.doc(
        responses={
            200: "Success",
            403: "Insufficient permissions",
            500: "Internal failure",
        }
    )
    def get(self, conf=None, server=None):
        """Reads the server configuration

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            {
              "boolean": [
                "daemon",
                "fork",
                "..."
              ],
              "defaults": {
                "address": "",
                "autoupgrade_dir": "",
                "ca_burp_ca": "",
                "ca_conf": "",
                "ca_name": "",
                "ca_server_name": "",
                "client_can_delete": true,
                "...": "..."
              },
              "integer": [
                "port",
                "status_port",
                "..."
              ],
              "multi": [
                "keep",
                "restore_client",
                "..."
              ],
              "placeholders": {
                "autoupgrade_dir": "path",
                "ca_burp_ca": "path",
                "ca_conf": "path",
                "ca_name": "name",
                "ca_server_name": "name",
                "client_can_delete": "0|1",
                "...": "..."
              },
              "results": {
                "boolean": [
                  {
                    "name": "hardlinked_archive",
                    "value": false,
                    "reset": false
                  },
                  {
                    "name": "syslog",
                    "value": true,
                    "reset": false
                  },
                  { "...": "..." }
                ],
                "clients": [
                  {
                    "name": "testclient",
                    "value": "/etc/burp/clientconfdir/testclient"
                  }
                ],
                "common": [
                  {
                    "name": "mode",
                    "value": "server",
                    "reset": false
                  },
                  {
                    "name": "directory",
                    "value": "/var/spool/burp",
                    "reset": false
                  },
                  { "...": "..." }
                ],
                "includes": [],
                "includes_ext": [],
                "integer": [
                  {
                    "name": "port",
                    "value": 4971,
                    "reset": false
                  },
                  {
                    "name": "status_port",
                    "value": 4972,
                    "reset": false
                  },
                  { "...": "..." }
                ],
                "multi": [
                  {
                    "name": "keep",
                    "value": [
                      "7",
                      "4"
                    ],
                    "reset": [
                      false,
                      true
                    ]
                  },
                  { "...": "..." }
                ],
                "hierarchy": [
                  {
                    "children": [
                      {
                        "children": [],
                        "dir": "/tmp/burp/conf.d",
                        "full": "/tmp/burp/conf.d/empty.conf",
                        "name": "empty.conf",
                        "parent": "/tmp/burp/burp-server.conf"
                      },
                      {
                        "children": [],
                        "dir": "/tmp/burp/conf.d",
                        "full": "/tmp/burp/conf.d/ipv4.conf",
                        "name": "ipv4.conf",
                        "parent": "/tmp/burp/burp-server.conf"
                      }
                    ],
                    "dir": "/tmp/burp",
                    "full": "/tmp/burp/burp-server.conf",
                    "name": "burp-server.conf",
                    "parent": null
                  }
                ]
              },
              "server_doc": {
                "address": "Defines the main TCP address...",
                "...": "..."
              },
              "string": [
                "mode",
                "address",
                "..."
              ],
              "suggest": {
                "compression": [
                  "gzip1",
                  "gzip2",
                  "gzip3",
                  "gzip4",
                  "gzip5",
                  "gzip6",
                  "gzip7",
                  "gzip8",
                  "gzip9"
                ],
                "mode": [
                  "client",
                  "server"
                ],
                "...": [
                  "..."
                ]
              }
            }


        :param server: Which server to collect data from when in multi-agent mode
        :type server: str

        :returns: The *JSON* described above.
        """
        try:
            conf = unquote(conf)
        except:
            pass
        res = bui.client.read_conf_srv(conf, server)
        refresh()
        # Translate the doc and placeholder API side
        cache_keys = {
            "doc": "_doc_parser_{}-{}".format(server, g.locale),
            "placeholders": "_placeholders_parser_{}-{}".format(server, g.locale),
            "boolean_srv": "_boolean_srv_parser_{}".format(server),
            "string_srv": "_string_srv_parser_{}".format(server),
            "integer_srv": "_integer_srv_parser_{}".format(server),
            "multi_srv": "_multi_srv_parser_{}".format(server),
            "values": "_suggest_parser_{}".format(server),
            "defaults": "_defaults_parser_{}".format(server),
            "advanced_type": "_advanced_parser_{}".format(server),
            "pair_associations": "_pair_associations_parser_{}".format(server),
        }
        cache_results = {}
        for name, key in cache_keys.items():
            if not cache.cache.has(key):
                if name in ["doc", "placeholders"]:
                    _tmp = bui.client.get_parser_attr(name, server).copy()
                    _tmp2 = {}
                    for k, v in _tmp.items():
                        _tmp2[k] = _(v)
                    cache_results[name] = _tmp2
                else:
                    cache_results[name] = bui.client.get_parser_attr(name, server)
                cache.cache.set(key, cache_results[name], 3600)
            else:
                cache_results[name] = cache.cache.get(key)

        return jsonify(
            results=res,
            boolean=cache_results["boolean_srv"],
            string=cache_results["string_srv"],
            integer=cache_results["integer_srv"],
            multi=cache_results["multi_srv"],
            pair=cache_results["pair_associations"],
            advanced=cache_results["advanced_type"],
            server_doc=cache_results["doc"],
            suggest=cache_results["values"],
            placeholders=cache_results["placeholders"],
            defaults=cache_results["defaults"],
        )


@ns.route("/clients", "/<server>/clients", endpoint="clients_list")
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
    },
)
class ClientsList(Resource):
    @api.acl_admin_or_moderator_required(
        message="Sorry, you don't have rights to access the setting panel"
    )
    @ns.doc(
        responses={
            200: "Success",
            403: "Insufficient permissions",
            500: "Internal failure",
        }
    )
    def get(self, server=None):
        """Returns a list of clients"""
        parser = bui.client.get_parser(agent=server)
        res = parser.list_clients()
        return jsonify(result=res)


@ns.route(
    "/static-templates", "/<server>/static-templates", endpoint="static_templates_list"
)
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
    },
)
class StaticTemplatesList(Resource):
    @api.acl_admin_or_moderator_required(
        message="Sorry, you don't have rights to access the setting panel"
    )
    @ns.doc(
        responses={
            200: "Success",
            403: "Insufficient permissions",
            500: "Internal failure",
        }
    )
    def get(self, server=None):
        """Returns a list of clients"""
        parser = bui.client.get_parser(agent=server)
        res = parser.list_static_templates()
        env = Environment()
        for obj in res:
            ast = env.parse(obj["content"])
            obj["variables"] = [
                x
                for x in meta.find_undeclared_variables(ast)
                if x not in TEMPLATE_EXCLUDES
            ]
        return jsonify(result=res)


@ns.route(
    "/static-template",
    "/<server>/static-template",
    endpoint="new_static_template",
    methods=["PUT"],
)
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
    },
)
class NewStaticTemplateSettings(Resource):
    parser = ns.parser()
    parser.add_argument(
        "newstatictemplate", required=True, help="No 'newstatictemplate' provided"
    )

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(
        message="Sorry, you don't have rights to access the setting panel"
    )
    @ns.expect(parser)
    @ns.doc(
        responses={
            200: "Success",
            400: "Missing parameter",
            403: "Insufficient permissions",
            500: "Internal failure",
        }
    )
    def put(self, server=None):
        """Creates a new template"""
        if (
            not current_user.is_anonymous
            and current_user.acl.is_moderator()
            and not current_user.acl.is_server_rw(server)
        ):
            self.abort(403, "You don't have rights on this server")

        newtemplate = self.parser.parse_args()["newstatictemplate"]
        if not newtemplate:
            self.abort(400, "No template name provided")
        parser = bui.client.get_parser(agent=server)
        templates = parser.list_static_templates()
        if any(tpl["name"] == newtemplate for tpl in templates):
            self.abort(409, "Static template '{}' already exists".format(newtemplate))
        # clientconfdir = bui.client.get_parser_attr('clientconfdir', server)
        # if not clientconfdir:
        #    flash('Could not proceed, no \'clientconfdir\' find', 'warning')
        #    return redirect(request.referrer)
        noti = bui.client.store_conf_cli(
            ImmutableMultiDict(), newtemplate, None, False, True, server
        )
        if server:
            url = url_for(
                "view.cli_settings",
                server=server,
                client=newtemplate,
                statictemplate=True,
            )
        else:
            url = url_for("view.cli_settings", client=newtemplate, statictemplate=True)
        noti.append(
            [
                NOTIF_INFO,
                _(
                    "<a href=\"%(url)s\">Click here</a> to edit '%(template)s' configuration",
                    url=url,
                    template=newtemplate,
                ),
            ]
        )
        # clear the cache when we add a new client
        cache.clear()
        bui.audit.logger.info(
            f"created new static template {newtemplate}", server=server
        )
        return {"notif": noti}, 201


@ns.route("/templates", "/<server>/templates", endpoint="templates_list")
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
    },
)
class TemplatesList(Resource):
    @api.acl_admin_or_moderator_required(
        message="Sorry, you don't have rights to access the setting panel"
    )
    @ns.doc(
        responses={
            200: "Success",
            403: "Insufficient permissions",
            500: "Internal failure",
        }
    )
    def get(self, server=None):
        """Returns a list of clients"""
        parser = bui.client.get_parser(agent=server)
        res = parser.list_templates()
        return jsonify(result=res)


@ns.route("/template", "/<server>/template", endpoint="new_template", methods=["PUT"])
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
    },
)
class NewTemplateSettings(Resource):
    parser = ns.parser()
    parser.add_argument("newtemplate", required=True, help="No 'newclient' provided")

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(
        message="Sorry, you don't have rights to access the setting panel"
    )
    @ns.expect(parser)
    @ns.doc(
        responses={
            200: "Success",
            400: "Missing parameter",
            403: "Insufficient permissions",
            500: "Internal failure",
        }
    )
    def put(self, server=None):
        """Creates a new template"""
        if (
            not current_user.is_anonymous
            and current_user.acl.is_moderator()
            and not current_user.acl.is_server_rw(server)
        ):
            self.abort(403, "You don't have rights on this server")

        newtemplate = self.parser.parse_args()["newtemplate"]
        if not newtemplate:
            self.abort(400, "No template name provided")
        parser = bui.client.get_parser(agent=server)
        templates = parser.list_templates()
        for tpl in templates:
            if tpl["name"] == newtemplate:
                self.abort(409, "Template '{}' already exists".format(newtemplate))
        # clientconfdir = bui.client.get_parser_attr('clientconfdir', server)
        # if not clientconfdir:
        #    flash('Could not proceed, no \'clientconfdir\' find', 'warning')
        #    return redirect(request.referrer)
        noti = bui.client.store_conf_cli(
            ImmutableMultiDict(), newtemplate, None, True, False, server
        )
        if server:
            url = url_for(
                "view.cli_settings", server=server, client=newtemplate, template=True
            )
        else:
            url = url_for("view.cli_settings", client=newtemplate, template=True)
        noti.append(
            [
                NOTIF_INFO,
                _(
                    "<a href=\"%(url)s\">Click here</a> to edit '%(template)s' configuration",
                    url=url,
                    template=newtemplate,
                ),
            ]
        )
        # clear the cache when we add a new client
        cache.clear()
        bui.audit.logger.info(f"created new template {newtemplate}", server=server)
        return {"notif": noti}, 201


@ns.route("/config", "/<server>/config", endpoint="new_client", methods=["PUT"])
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
    },
)
class NewClientSettings(Resource):
    parser = ns.parser()
    parser.add_argument("newclient", required=True, help="No 'newclient' provided")
    parser.add_argument("templates", help="Templates list", action="split")
    parser.add_argument("statictemplate", help="Static template")
    parser.add_argument("variables", help="Template variables")

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(
        message="Sorry, you don't have rights to access the setting panel"
    )
    @ns.expect(parser)
    @ns.doc(
        responses={
            200: "Success",
            400: "Missing parameter",
            403: "Insufficient permissions",
            500: "Internal failure",
        }
    )
    def put(self, server=None):
        """Creates a new client"""
        args = self.parser.parse_args()
        newclient = args["newclient"]
        templates = [x for x in args.get("templates", []) if x]
        statictemplate = args["statictemplate"]
        variables = json.loads(args["variables"]) if args["variables"] else {}
        variables["agent"] = server
        variables["client"] = newclient
        if not newclient:
            self.abort(400, "No client name provided")

        if (
            not current_user.is_anonymous
            and current_user.acl.is_moderator()
            and not current_user.acl.is_client_rw(newclient, server)
        ):
            self.abort(403, "You don't have rights on this server")

        parser = bui.client.get_parser(agent=server)
        clients = parser.list_clients()
        for cl in clients:
            if cl["name"] == newclient:
                self.abort(409, "Client '{}' already exists".format(newclient))
        # clientconfdir = bui.client.get_parser_attr('clientconfdir', server)
        # if not clientconfdir:
        #    flash('Could not proceed, no \'clientconfdir\' find', 'warning')
        #    return redirect(request.referrer)
        data = MultiDict()
        content = ""
        if templates:
            real_templates = {x["name"]: x["value"] for x in parser._list_templates()}
            if any(x not in real_templates for x in templates):
                self.abort(400, "Wrong template")
            data.setlist("templates", [real_templates[x] for x in templates])
        if statictemplate:
            statics = parser._list_static_templates()
            for tpl in statics:
                if tpl["name"] == statictemplate:
                    content = render_template_string(tpl["content"], **variables)
        noti = bui.client.store_conf_cli(
            ImmutableMultiDict(data), newclient, None, content=content, agent=server
        )
        if server:
            url = url_for("view.cli_settings", server=server, client=newclient)
        else:
            url = url_for("view.cli_settings", client=newclient)
        noti.append(
            [
                NOTIF_INFO,
                _(
                    "<a href=\"%(url)s\">Click here</a> to edit '%(client)s' configuration",
                    url=url,
                    client=newclient,
                ),
            ]
        )
        # clear the cache when we add a new client
        cache.clear()
        # clear client-side cache through the _extra META variable
        try:
            _extra = session.get("_extra", g.now)
            _extra = int(_extra)
        except ValueError:
            _extra = 0
        session["_extra"] = "{}".format(_extra + 1)
        if bui.config["WITH_CELERY"]:
            from ..tasks import force_scheduling_now

            force_scheduling_now()

        bui.audit.logger.info(
            f"created new client configuration {newclient}", server=server
        )
        return {"notif": noti}, 201


@ns.route(
    "/config/<client>",
    "/config/<client>/<path:conf>",
    "/<server>/config/<client>",
    "/<server>/config/<client>/<path:conf>",
    endpoint="client_settings",
    methods=["GET", "POST", "PUT", "DELETE"],
)
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
        "client": "Client name",
        "conf": "Path of the configuration file",
    },
)
class ClientSettings(Resource):
    parser_delete = ns.parser()
    parser_delete.add_argument(
        "revoke",
        type=inputs.boolean,
        help="Whether to revoke the certificate or not",
        default=False,
        nullable=True,
    )
    parser_delete.add_argument(
        "delcert",
        type=inputs.boolean,
        help="Whether to delete the certificate or not",
        default=False,
        nullable=True,
    )
    parser_delete.add_argument(
        "keepconf",
        type=inputs.boolean,
        help="Whether to keep the conf or not",
        default=False,
        nullable=True,
    )
    parser_delete.add_argument(
        "template",
        type=inputs.boolean,
        help="Whether we work on a template or not",
        default=False,
        nullable=True,
    )
    parser_delete.add_argument(
        "statictemplate",
        type=inputs.boolean,
        help="Whether we work on a static template or not",
        default=False,
        nullable=True,
    )
    parser_delete.add_argument(
        "delete",
        type=inputs.boolean,
        help="Whether we should remove the data as well or not",
        default=False,
        nullable=True,
    )
    parser_put = ns.parser()
    parser_put.add_argument("newname", help="New name of the client/template")
    parser_put.add_argument(
        "template",
        type=inputs.boolean,
        help="Whether we work on a template or not",
        default=False,
        nullable=True,
    )
    parser_put.add_argument(
        "statictemplate",
        type=inputs.boolean,
        help="Whether we work on a static template or not",
        default=False,
        nullable=True,
    )
    parser_put.add_argument(
        "keepcert",
        type=inputs.boolean,
        help="Whether to keep the same certificate or not",
        default=False,
        nullable=True,
    )
    parser_put.add_argument(
        "keepdata",
        type=inputs.boolean,
        help="Whether to keep the data or not",
        default=False,
        nullable=True,
    )
    parser_post = ns.parser()
    parser_post.add_argument(
        "template",
        type=inputs.boolean,
        help="Whether we work on a template or not",
        default=False,
        nullable=True,
    )
    parser_post.add_argument(
        "statictemplate",
        type=inputs.boolean,
        help="Whether we work on a static template or not",
        default=False,
        nullable=True,
    )
    parser_get = ns.parser()
    parser_get.add_argument(
        "template",
        type=inputs.boolean,
        help="Whether we work on a template or not",
        default=False,
        nullable=True,
    )
    parser_get.add_argument(
        "statictemplate",
        type=inputs.boolean,
        help="Whether we work on a static template or not",
        default=False,
        nullable=True,
    )

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(
        message=_("Sorry, you don't have rights to access the setting panel")
    )
    @ns.expect(parser_post)
    @ns.doc(
        responses={
            200: "Success",
            403: "Insufficient permissions",
            500: "Internal failure",
        }
    )
    def post(self, server=None, client=None, conf=None):
        """Saves a given client configuration"""
        if (
            not current_user.is_anonymous
            and current_user.acl.is_moderator()
            and not current_user.acl.is_client_rw(client, server)
        ):
            self.abort(403, "You don't have rights on this server")

        args = self.parser_post.parse_args()
        template = args.get("template", False)
        statictemplate = args.get("statictemplate", False)
        noti = bui.client.store_conf_cli(
            request.form, client, conf, template, statictemplate, server
        )
        # clear cache
        cache.clear()
        # clear client-side cache through the _extra META variable
        try:
            _extra = session.get("_extra", g.now)
            _extra = int(_extra)
        except ValueError:
            _extra = 0
        session["_extra"] = "{}".format(_extra + 1)

        bui.audit.logger.info(
            f"updated client configuration {client} ({conf})", server=server
        )
        return {"notif": noti}

    @api.acl_admin_or_moderator_required(
        message=_("Sorry, you don't have rights to access the setting panel")
    )
    @ns.expect(parser_get)
    @ns.doc(
        responses={
            200: "Success",
            403: "Insufficient permissions",
            500: "Internal failure",
        }
    )
    def get(self, server=None, client=None, conf=None):
        """Reads a given client configuration"""
        try:
            conf = unquote(conf)
        except:
            pass
        args = self.parser_get.parse_args()
        template = args.get("template", False)
        statictemplate = args.get("statictemplate", False)
        parser = bui.client.get_parser(agent=server)
        res = parser.read_client_conf(client, conf, template, statictemplate)
        refresh()
        # Translate the doc and placeholder API side
        cache_keys = {
            "doc": "_doc_parser_{}-{}".format(server, g.locale),
            "placeholders": "_placeholders_parser_{}-{}".format(server, g.locale),
            "boolean_cli": "_boolean_cli_parser_{}".format(server),
            "string_cli": "_string_cli_parser_{}".format(server),
            "integer_cli": "_integer_cli_parser_{}".format(server),
            "multi_cli": "_multi_cli_parser_{}".format(server),
            "values": "_suggest_parser_{}".format(server),
            "defaults": "_defaults_parser_{}".format(server),
        }
        cache_results = {}
        for name, key in cache_keys.items():
            if not cache.cache.has(key):
                if name in ["doc", "placeholders"]:
                    _tmp = bui.client.get_parser_attr(name, server).copy()
                    _tmp2 = {}
                    for k, v in _tmp.items():
                        _tmp2[k] = _(v)
                    cache_results[name] = _tmp2
                else:
                    cache_results[name] = bui.client.get_parser_attr(name, server)
                cache.cache.set(key, cache_results[name], 3600)
            else:
                cache_results[name] = cache.cache.get(key)

        return jsonify(
            results=res,
            boolean=cache_results["boolean_cli"],
            string=cache_results["string_cli"],
            integer=cache_results["integer_cli"],
            multi=cache_results["multi_cli"],
            server_doc=cache_results["doc"],
            suggest=cache_results["values"],
            placeholders=cache_results["placeholders"],
            defaults=cache_results["defaults"],
        )

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(
        message=_("Sorry, you don't have rights to access the setting panel")
    )
    @ns.expect(parser_delete)
    @ns.doc(
        responses={
            200: "Success",
            403: "Insufficient permissions",
            409: "Conflict",
            500: "Internal failure",
        }
    )
    def delete(self, server=None, client=None, conf=None):
        """Deletes a given client"""
        if (
            not current_user.is_anonymous
            and current_user.acl.is_moderator()
            and not current_user.acl.is_client_rw(client, server)
        ):
            self.abort(403, "You don't have rights on this server")

        if bui.client.is_backup_running(client, server):
            self.abort(
                409,
                "There is currently a backup running for this client hence "
                "we cannot delete it for now. Please try again later",
            )

        args = self.parser_delete.parse_args()
        delcert = args.get("delcert", False)
        revoke = args.get("revoke", False)
        keepconf = args.get("keepconf", False)
        template = args.get("template", False)
        statictemplate = args.get("statictemplate", False)
        delete = args.get("delete", False)

        if not keepconf:
            # clear the cache when we remove a client
            cache.clear()
            # clear client-side cache through the _extra META variable
            try:
                _extra = session.get("_extra", g.now)
                _extra = int(_extra)
            except ValueError:
                _extra = 0
            session["_extra"] = "{}".format(_extra + 1)
            if bui.config["WITH_CELERY"]:
                from ..tasks import force_scheduling_now

                force_scheduling_now()
        parser = bui.client.get_parser(agent=server)

        bui.audit.logger.info(
            f"deleted client configuration {client}, delete certificate: {delcert}, "
            f"revoke certificate: {revoke}, keep a backup of the configuration: "
            f"{keepconf}, delete data: {delete}, is template: {template} "
            f"is static template: {statictemplate}",
            server=server,
        )
        return (
            parser.remove_client(
                client, keepconf, delcert, revoke, template, statictemplate, delete
            ),
            200,
        )

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(
        message=_("Sorry, you don't have rights to access the setting panel")
    )
    @ns.expect(parser_put)
    @ns.doc(
        responses={
            200: "Success",
            403: "Insufficient permissions",
            409: "Conflict",
            500: "Internal failure",
        }
    )
    def put(self, server=None, client=None, conf=None):
        """Rename a given client"""
        if (
            not current_user.is_anonymous
            and current_user.acl.is_moderator()
            and not current_user.acl.is_client_rw(client, server)
        ):
            self.abort(403, "You don't have rights on this server")

        if bui.client.is_backup_running(client, server):
            self.abort(
                409,
                "There is currently a backup running for this client hence "
                "we cannot delete it for now. Please try again later",
            )

        args = self.parser_put.parse_args()
        newname = args.get("newname", None)
        keepcert = args.get("keepcert", False)
        keepdata = args.get("keepdata", False)
        template = args.get("template", False)
        statictemplate = args.get("statictemplate", False)

        # clear the cache when we remove a client
        cache.clear()
        # clear client-side cache through the _extra META variable
        try:
            _extra = session.get("_extra", g.now)
            _extra = int(_extra)
        except ValueError:
            _extra = 0
        session["_extra"] = "{}".format(_extra + 1)
        if bui.config["WITH_CELERY"]:
            from ..tasks import force_scheduling_now

            force_scheduling_now()
        parser = bui.client.get_parser(agent=server)

        bui.audit.logger.info(
            f"renaming client configuration {client} to {newname}, "
            f"keep data: {keepdata}, keep certificate: {keepcert}, "
            f"is template: {template}, is static template: {statictemplate}",
            server=server,
        )
        return (
            parser.rename_client(
                client, newname, template, statictemplate, keepcert, keepdata
            ),
            200,
        )


@ns.route(
    "/path-expander",
    "/<server>/path-expander",
    "/path-expander/<client>",
    "/<server>/path-expander/<client>",
    endpoint="path_expander",
)
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
        "client": "Client name",
    },
)
class PathExpander(Resource):

    parser = ns.parser()
    parser.add_argument("path", required=True, help="No 'path' provided")
    parser.add_argument("source", required=False, help="Which file is it included in")

    @api.acl_admin_or_moderator_required(
        message=_("Sorry, you don't have rights to access the setting panel")
    )
    @ns.doc(
        responses={
            200: "Success",
            403: "Insufficient permissions",
            500: "Internal failure",
        }
    )
    def get(self, server=None, client=None):
        """Expends a given path

        For instance if it's given a glob expression it will returns a list of
        files matching the expression.
        """
        args = self.parser.parse_args()
        path = args["path"]
        source = args["source"]
        if path:
            path = unquote(path)
        if source:
            source = unquote(source)
        parser = bui.client.get_parser(agent=server)
        paths = parser.path_expander(path, source, client)
        if not paths:
            self.abort(403, "Path not found")
        return {"result": paths}


@ns.route("/options", "/<server>/options", endpoint="setting_options")
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
    },
)
class SettingOptions(Resource):
    @ns.doc(
        responses={
            200: "Success",
            403: "Insufficient permissions",
            500: "Internal failure",
        }
    )
    def get(self, server=None):
        """Returns various setting options"""
        return {
            "is_revocation_enabled": bui.client.revocation_enabled(server),
            "server_can_restore": not bui.noserverrestore
            or bui.client.get_parser(agent=server).param(
                "server_can_restore", "client_conf"
            ),
            "batch_list_supported": bui.client.get_attr(
                "batch_list_supported", False, server
            ),
        }



# -*- coding: utf8 -*-
"""
.. module:: burpui.api.admin
    :platform: Unix
    :synopsis: Burp-UI admin api module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from . import api
from ..engines.server import BUIServer  # noqa
from ..sessions import session_manager
from ..misc.acl.meta import meta_grants
from ..utils import NOTIF_OK
from .custom import fields, Resource

#  from ..exceptions import BUIserverException

from flask import current_app
from flask_login import current_user
from flask_babel import gettext

import json

bui = current_app  # type: BUIServer
ns = api.namespace("admin", "Admin methods")

user_fields = ns.model(
    "Users",
    {
        "id": fields.String(required=True, description="User id"),
        "name": fields.String(required=True, description="User name"),
        "backend": fields.String(required=True, description="Backend name"),
    },
)
grant_fields = ns.model(
    "Grants",
    {
        "id": fields.String(required=True, description="Grant id"),
        "grant": fields.String(required=True, description="Grant content"),
        "backend": fields.String(required=True, description="Backend name"),
    },
)
group_fields = ns.model(
    "Groups",
    {
        "id": fields.String(required=True, description="Group id"),
        "grant": fields.String(required=True, description="Group grant content"),
        "members": fields.List(
            fields.String, required=True, description="Group members"
        ),
        "backend": fields.String(required=True, description="Backend name"),
    },
)
groups_fields = ns.model(
    "GroupsFields",
    {
        "name": fields.String(required=True, description="Group name"),
        "inherit": fields.List(
            fields.String, required=False, description="This group is inherited by"
        ),
    },
)
groups_list_fields = ns.model(
    "GroupsList",
    {
        "groups": fields.List(
            fields.Nested(groups_fields), required=True, description="Groups list"
        ),
    },
)
group_members_fields = ns.model(
    "GroupMembers",
    {
        "members": fields.List(
            fields.String, required=True, description="Group members"
        ),
        "grant": fields.String(required=True, description="Group grant content"),
    },
)
is_moderator_fields = ns.model(
    "IsModerator",
    {
        "moderator": fields.Boolean(
            required=True, description="Is the member a moderator"
        ),
        "inherit": fields.List(
            fields.String,
            required=False,
            description="What provides this grant if inherited",
        ),
    },
)
moderator_members_fields = ns.model(
    "ModeratorMembers",
    {
        "members": fields.List(
            fields.String, required=True, description="Moderator members"
        ),
        "grant": fields.String(required=True, description="Moderator grant content"),
    },
)
moderators_fields = ns.model(
    "Moderators",
    {
        "members": fields.List(
            fields.String, required=True, description="Moderator members"
        ),
        "grant": fields.String(required=True, description="Moderator grant content"),
        "backend": fields.String(required=True, description="Backend name"),
    },
)
is_admin_fields = ns.model(
    "IsAdmin",
    {
        "admin": fields.Boolean(required=True, description="Is the member an admin"),
        "inherit": fields.List(
            fields.String,
            required=False,
            description="What provides this grant if inherited",
        ),
    },
)
admin_members_fields = ns.model(
    "AdminMembers",
    {
        "members": fields.List(
            fields.String, required=True, description="Admin members"
        ),
    },
)
admins_fields = ns.model(
    "Admins",
    {
        "members": fields.List(
            fields.String, required=True, description="Admin members"
        ),
        "backend": fields.String(required=True, description="Backend name"),
    },
)
session_fields = ns.model(
    "Sessions",
    {
        "uuid": fields.String(description="Session id"),
        "ip": fields.String(description="IP address"),
        "ua": fields.String(description="User-Agent"),
        "permanent": fields.Boolean(description="Remember cookie"),
        "api": fields.Boolean(description="API login"),
        "expire": fields.DateTime(description="Expiration date"),
        "timestamp": fields.DateTime(description="Last seen"),
        "current": fields.Boolean(description="Is current session", default=False),
    },
)
acl_backend_fields = ns.model(
    "AclBackends",
    {
        "name": fields.String(required=True, description="Backend name"),
        "description": fields.String(required=True, description="Backend description"),
        "type": fields.String(required=False, description="Backend type"),
        "priority": fields.Integer(required=False, description="Backend priority"),
        "add_grant": fields.Boolean(
            required=False, default=False, description="Support grant creation"
        ),
        "mod_grant": fields.Boolean(
            required=False, default=False, description="Support grant edition"
        ),
        "del_grant": fields.Boolean(
            required=False, default=False, description="Support grant deletion"
        ),
        "add_group": fields.Boolean(
            required=False, default=False, description="Support group creation"
        ),
        "mod_group": fields.Boolean(
            required=False, default=False, description="Support group edition"
        ),
        "del_group": fields.Boolean(
            required=False, default=False, description="Support group deletion"
        ),
        "add_group_member": fields.Boolean(
            required=False, default=False, description="Support group member addition"
        ),
        "del_group_member": fields.Boolean(
            required=False, default=False, description="Support group member deletion"
        ),
        "add_moderator": fields.Boolean(
            required=False, default=False, description="Support moderator creation"
        ),
        "mod_moderator": fields.Boolean(
            required=False, default=False, description="Support moderator edition"
        ),
        "del_moderator": fields.Boolean(
            required=False, default=False, description="Support moderator deletion"
        ),
        "add_admin": fields.Boolean(
            required=False, default=False, description="Support admin creation"
        ),
        "del_admin": fields.Boolean(
            required=False, default=False, description="Support admin deletion"
        ),
    },
)
auth_backend_fields = ns.model(
    "Backends",
    {
        "name": fields.String(required=True, description="Backend name"),
        "description": fields.String(required=True, description="Backend description"),
        "type": fields.String(required=False, description="Backend type"),
        "priority": fields.Integer(required=False, description="Backend priority"),
        "add": fields.Boolean(
            required=False, default=False, description="Support user creation"
        ),
        "mod": fields.Boolean(
            required=False, default=False, description="Support user edition"
        ),
        "del": fields.Boolean(
            required=False, default=False, description="Support user deletion"
        ),
    },
)


@ns.route("/me", endpoint="admin_me")
class AdminMe(Resource):
    """The :class:`burpui.api.admin.AdminMe` resource allows you to
    retrieve informations about your currently logged in user.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    @ns.marshal_with(user_fields)
    def get(self):
        """Returns the current user informations

        **GET** method provided by the webservice.

        :returns: User
        """
        ret = getattr(current_user, "real", current_user)
        return ret


@ns.route(
    "/acl/isAdmin/<member>", "/acl/<backend>/isAdmin/<member>", endpoint="acl_is_admin"
)
@ns.doc(
    params={
        "member": "Username",
    }
)
class AclIsAdmin(Resource):
    """The :class:`burpui.api.admin.AclIsAdmin` resources allows you to check
    if a given member is admin or not.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    @api.acl_admin_or_moderator_required(message="Not allowed to view admins list")
    @ns.marshal_with(is_admin_fields)
    @ns.doc(
        responses={
            403: "Insufficient permissions",
            404: "No backend found",
        },
    )
    def get(self, member, backend=None):
        """Checks if a given member is admin"""
        if not backend:
            (ret, inh) = meta_grants.is_admin(member)
            return {"admin": ret, "inherit": inh}
        try:
            handler = getattr(bui, "acl_handler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        if backend not in handler.backends:
            self.abort(404, "No acl backend '{}' found".format(backend))
        loader = handler.backends[backend]
        return {"admin": member in loader.admins}


@ns.route("/acl/admins", "/acl/<backend>/admins", endpoint="acl_admins")
@ns.doc(
    params={
        "backend": "Backend name",
    }
)
class AclAdminss(Resource):
    """The :class:`burpui.api.admin.AclAdminss` resource allows you to
    retrieve a list of admins.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    @api.acl_admin_or_moderator_required(message="Not allowed to view groups list")
    @ns.marshal_list_with(admins_fields, code=200, description="Success")
    @ns.doc(
        responses={
            403: "Insufficient permissions",
            404: "No backend found",
        },
    )
    def get(self, name=None, backend=None):
        """Returns a list of admins

        **GET** method provided by the webservice.

        :returns: Moderators
        """
        try:
            handler = getattr(bui, "acl_handler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        ret = []
        for _, loader in handler.backends.items():
            append = {"members": loader.admins, "backend": loader.name}
            if (backend and backend == append["backend"]) or backend is None:
                return [append]
            ret.append(append)
        return ret


@ns.route(
    "/acl/admin",
    "/acl/<backend>/admin",
    "/acl/<backend>/admin/<member>",
    endpoint="acl_admin",
)
@ns.doc(
    params={
        "backend": "ACL backend",
        "member": "Admin member",
    }
)
class AclAdmin(Resource):
    """The :class:`burpui.api.admin.AclAdmins` resource allows you to
    retrieve a list of admins and add/delete them if your
    acl backend support those actions.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    parser = ns.parser()
    parser.add_argument(
        "memberNames", required=False, help="Admin members", action="append"
    )
    parser.add_argument("backendName", required=False, help="Backend name")

    @api.acl_admin_or_moderator_required(message="Not allowed to view admins list")
    @ns.marshal_with(admin_members_fields, code=200, description="Success")
    @ns.doc(
        responses={
            403: "Insufficient permissions",
            404: "No backend found",
        },
    )
    def get(self, backend):
        """Returns a list of admin users

        **GET** method provided by the webservice.

        :returns: Members
        """
        try:
            handler = getattr(bui, "acl_handler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        if backend not in handler.backends:
            self.abort(404, "No acl backend '{}' found".format(backend))
        ret = {}
        loader = handler.backends[backend]
        ret = {"members": loader.admins}
        return ret

    @api.disabled_on_demo()
    @api.acl_admin_required(message="Not allowed to add admin members")
    @ns.expect(parser)
    @ns.doc(
        responses={
            403: "Insufficient permissions",
            404: "No backend found",
        },
    )
    def put(self, backend=None, member=None):
        """Add a member as admin

        **PUT** method provided by the webservice.
        """
        args = self.parser.parse_args()
        backend = backend or args["backendName"]
        try:
            handler = getattr(bui, "acl_handler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        if backend not in handler.backends:
            self.abort(404, "No acl backend '{}' found".format(backend))

        loader = handler.backends[backend]

        members = [member] if member else (args["memberNames"] or [])

        if loader.add_admin is False:
            self.abort(
                500,
                "The '{}' backend does not support moderator member addition"
                "".format(backend),
            )

        ret = []
        status = 200

        for member in members:
            success, message, code = loader.add_admin(member)
            status = 201 if success else 200
            ret.append([code, message])

        bui.audit.logger.info(f"granted {members} as admin")
        return ret, status

    @api.disabled_on_demo()
    @api.acl_admin_required(message="Not allowed to remove admin members")
    @ns.expect(parser)
    @ns.doc(
        responses={
            403: "Insufficient permissions",
            404: "No backend found",
        },
    )
    def delete(self, backend=None, member=None):
        """Remove an admin member

        **DELETE** method provided by the webservice.
        """
        args = self.parser.parse_args()
        backend = backend or args["backendName"]
        try:
            handler = getattr(bui, "acl_handler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        if backend not in handler.backends:
            self.abort(40422, "No acl backend '{}' found".format(backend))

        loader = handler.backends[backend]

        members = [member] if member else (args["memberNames"] or [])

        if loader.del_admin is False:
            self.abort(
                500,
                "The '{}' backend does not support admin member deletion"
                "".format(backend),
            )

        ret = []
        status = 200

        for member in members:
            success, message, code = loader.del_admin(member)
            status = 201 if success else 200
            ret.append([code, message])

        bui.audit.logger.info(f"removed admin grants of {members}")
        return ret, status


@ns.route(
    "/acl/isModerator/<member>",
    "/acl/<backend>/isModerator/<member>",
    endpoint="acl_is_moderator",
)
@ns.doc(
    params={
        "member": "Username",
    }
)
class AclIsModerator(Resource):
    """The :class:`burpui.api.admin.AclIsModerator` resources allows you to check
    if a given member is moderator or not.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    @api.acl_admin_or_moderator_required(message="Not allowed to view admins list")
    @ns.marshal_with(is_moderator_fields)
    @ns.doc(
        responses={
            403: "Insufficient permissions",
            404: "No backend found",
        },
    )
    def get(self, member, backend=None):
        """Checks if a given member is moderator"""
        if not backend:
            (ret, inh) = meta_grants.is_moderator(member)
            return {"moderator": ret, "inherit": inh}
        try:
            handler = getattr(bui, "acl_handler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        if backend not in handler.backends:
            self.abort(404, "No acl backend '{}' found".format(backend))
        loader = handler.backends[backend]
        return {"moderator": member in loader.moderators}


@ns.route("/acl/moderators", "/acl/<backend>/moderators", endpoint="acl_moderators")
@ns.doc(
    params={
        "backend": "Backend name",
    }
)
class AclModerators(Resource):
    """The :class:`burpui.api.admin.AclModerators` resource allows you to
    retrieve a list of moderators.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    @api.acl_admin_or_moderator_required(message="Not allowed to view groups list")
    @ns.marshal_list_with(moderators_fields, code=200, description="Success")
    @ns.doc(
        responses={
            403: "Insufficient permissions",
            404: "No backend found",
        },
    )
    def get(self, name=None, backend=None):
        """Returns a list of moderators

        **GET** method provided by the webservice.

        :returns: Moderators
        """
        try:
            handler = getattr(bui, "acl_handler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        ret = []
        for _, loader in handler.backends.items():
            append = {
                "grant": loader.moderator,
                "members": loader.moderators,
                "backend": loader.name,
            }
            if (backend and backend == append["backend"]) or backend is None:
                return [append]
            ret.append(append)
        return ret


@ns.route(
    "/acl/moderator",
    "/acl/<backend>/moderator",
    "/acl/<backend>/moderator/<member>",
    endpoint="acl_moderator",
)
@ns.doc(
    params={
        "backend": "ACL backend",
        "member": "Moderator member",
    }
)
class AclModerator(Resource):
    """The :class:`burpui.api.admin.AclModerator` resource allows you to
    retrieve a list of moderators and add/delete them if your
    acl backend support those actions.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    parser = ns.parser()
    parser.add_argument(
        "memberNames", required=False, help="Moderator members", action="append"
    )
    parser.add_argument("backendName", required=False, help="Backend name")

    parser_mod = ns.parser()
    parser_mod.add_argument("grant", required=False, help="Moderator grants")

    @api.acl_admin_or_moderator_required(message="Not allowed to view moderators list")
    @ns.marshal_with(moderator_members_fields, code=200, description="Success")
    @ns.doc(
        responses={
            403: "Insufficient permissions",
            404: "No backend found",
        },
    )
    def get(self, backend):
        """Returns a list of moderator users

        **GET** method provided by the webservice.

        :returns: Members
        """
        try:
            handler = getattr(bui, "acl_handler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        if backend not in handler.backends:
            self.abort(404, "No acl backend '{}' found".format(backend))
        ret = {}
        loader = handler.backends[backend]
        ret = {"members": loader.moderators, "grant": loader.moderator}
        return ret

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(message="Not allowed to add moderator members")
    @ns.expect(parser)
    @ns.doc(
        responses={
            403: "Insufficient permissions",
            404: "No backend found",
        },
    )
    def put(self, backend=None, member=None):
        """Add a member as moderator

        **PUT** method provided by the webservice.
        """
        args = self.parser.parse_args()
        backend = backend or args["backendName"]
        try:
            handler = getattr(bui, "acl_handler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        if backend not in handler.backends:
            self.abort(404, "No acl backend '{}' found".format(backend))

        loader = handler.backends[backend]

        members = [member] if member else (args["memberNames"] or [])

        if loader.add_moderator is False:
            self.abort(
                500,
                "The '{}' backend does not support moderator member addition"
                "".format(backend),
            )

        ret = []
        status = 200
        for member in members:
            success, message, code = loader.add_moderator(member)
            ret.append([code, message])
            status = 201 if success else 200

        bui.audit.logger.info(f"granted {members} as moderator")
        return ret, status

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(
        message="Not allowed to remove moderator members"
    )
    @ns.expect(parser)
    @ns.doc(
        responses={
            403: "Insufficient permissions",
            404: "No backend found",
        },
    )
    def delete(self, backend=None, member=None):
        """Remove a moderator member

        **DELETE** method provided by the webservice.
        """
        args = self.parser.parse_args()
        backend = backend or args["backendName"]
        try:
            handler = getattr(bui, "acl_handler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        if backend not in handler.backends:
            self.abort(404, "No acl backend '{}' found".format(backend))

        loader = handler.backends[backend]

        members = [member] if member else (args["memberNames"] or [])

        if loader.del_moderator is False:
            self.abort(
                500,
                "The '{}' backend does not support moderator member deletion"
                "".format(backend),
            )

        ret = []
        status = 200
        for member in members:
            success, message, code = loader.del_moderator(member)
            ret.append([code, message])
            status = 201 if success else 200

        bui.audit.logger.info(f"removed moderator grant of {members}")
        return ret, status

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(
        message="Not allowed to update moderator grants"
    )
    @ns.expect(parser_mod)
    @ns.doc(
        responses={
            403: "Insufficient permissions",
            404: "No backend found",
        },
    )
    def post(self, backend):
        """Update moderator grants

        **POST** method provided by the webservice.
        """
        args = self.parser_mod.parse_args()
        try:
            handler = getattr(bui, "acl_handler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        if backend not in handler.backends:
            self.abort(404, "No acl backend '{}' found".format(backend))

        loader = handler.backends[backend]

        grants = args["grant"]

        if loader.mod_moderator is False:
            self.abort(
                500,
                "The '{}' backend does not support moderator grants edition"
                "".format(backend),
            )

        success, message, code = loader.mod_moderator(grants)
        status = 201 if success else 200

        bui.audit.logger.info(f"updated moderator grants to: {grants}")
        return [[code, message]], status


@ns.route(
    "/acl/group",
    "/acl/<backend>/group/<name>",
    "/acl/<backend>/group/<name>/<member>",
    endpoint="acl_group_members",
)
@ns.doc(
    params={
        "name": "Group name",
        "backend": "ACL backend",
        "member": "Group member",
    }
)
class AclGroup(Resource):
    """The :class:`burpui.api.admin.AclGroup` resource allows you to
    retrieve a list of members in a given group and add/delete them if your
    acl backend support those actions.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    parser = ns.parser()
    parser.add_argument(
        "memberNames", required=False, help="Group members", action="append"
    )
    parser.add_argument("groupName", required=False, help="Group name")
    parser.add_argument("backendName", required=False, help="Backend name")

    @api.acl_admin_or_moderator_required(message="Not allowed to view groups list")
    @ns.marshal_with(group_members_fields, code=200, description="Success")
    @ns.doc(
        responses={
            403: "Insufficient permissions",
            404: "No backend found",
        },
    )
    def get(self, name, backend):
        """Returns a list of users in a giver group

        **GET** method provided by the webservice.

        :returns: Members
        """
        try:
            handler = getattr(bui, "acl_handler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        if backend not in handler.backends:
            self.abort(404, "No acl backend '{}' found".format(backend))
        ret = {}
        loader = handler.backends[backend]
        groups = loader.groups
        gname = "@{}".format(name)
        if groups and gname in groups:
            ret = {
                "members": groups[gname].get("members", []),
                "grant": groups[gname].get("grants", ""),
            }
        return ret

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(message="Not allowed to add member in group")
    @ns.expect(parser)
    @ns.doc(
        responses={
            403: "Insufficient permissions",
            404: "No backend found",
        },
    )
    def put(self, name=None, backend=None, member=None):
        """Add a member in a given group

        **PUT** method provided by the webservice.
        """
        args = self.parser.parse_args()
        name = name or args["groupName"]
        backend = backend or args["backendName"]
        try:
            handler = getattr(bui, "acl_handler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        if backend not in handler.backends:
            self.abort(404, "No acl backend '{}' found".format(backend))

        loader = handler.backends[backend]

        members = [member] if member else (args["memberNames"] or [])

        if loader.add_group_member is False:
            self.abort(
                500,
                "The '{}' backend does not support group member addition"
                "".format(backend),
            )

        ret = []
        status = 200
        for member in members:
            success, message, code = loader.add_group_member(name, member)
            ret.append([code, message])
            status = 201 if success else 200

        bui.audit.logger.info(f"added {members} in group {name}")
        return ret, status

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(
        message="Not allowed to remove member in group"
    )
    @ns.expect(parser)
    @ns.doc(
        responses={
            403: "Insufficient permissions",
            404: "No backend found",
        },
    )
    def delete(self, name, backend, member=None):
        """Remove a member from a given group

        **DELETE** method provided by the webservice.
        """
        args = self.parser.parse_args()
        try:
            handler = getattr(bui, "acl_handler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        if backend not in handler.backends:
            self.abort(404, "No acl backend '{}' found".format(backend))

        loader = handler.backends[backend]

        members = [member] if member else (args["memberNames"] or [])

        if loader.del_group_member is False:
            self.abort(
                500,
                "The '{}' backend does not support group member deletion"
                "".format(backend),
            )

        ret = []
        status = 200
        for member in members:
            success, message, code = loader.del_group_member(name, member)
            ret.append([code, message])
            status = 201 if success else 200

        bui.audit.logger.info(f"removed {members} from group {name}")
        return ret, status


@ns.route("/acl/groupsOf/<member>", endpoint="acl_groups_of")
@ns.doc(
    params={
        "member": "Username",
    }
)
class AclGroupsOf(Resource):
    """The :class:`burpui.api.admin.AclGroupsOf` resource allows you to retrieve
    a list of groups of a given user.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    @api.acl_admin_or_moderator_required(message="Not allowed to view groups list")
    @ns.marshal_with(groups_list_fields, code=200, description="Success")
    @ns.doc(
        responses={
            403: "Insufficient permissions",
            404: "No backend found",
        },
    )
    def get(self, member):
        """Returns a list of group

        **GET** method provided by the webservice.

        :returns: Groups
        """
        return {
            "groups": [
                {"name": name, "inherit": inherit}
                for name, inherit in meta_grants.get_member_groups(member)
            ]
        }


@ns.route(
    "/acl/groups",
    "/acl/<backend>/groups",
    "/acl/groups/<name>",
    "/acl/<backend>/groups/<name>",
    endpoint="acl_groups",
)
@ns.doc(
    params={
        "name": "Group name",
        "backend": "Backend name",
    }
)
class AclGroups(Resource):
    """The :class:`burpui.api.admin.AclGroups` resource allows you to
    retrieve a list of groups and to add/update/delete them if your
    acl backend support those actions.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    parser_add = ns.parser()
    parser_add.add_argument(
        "group", required=True, help="Group name", location="values"
    )
    parser_add.add_argument("grant", required=True, help="Group grant content")
    parser_add.add_argument("backend", help="Backend", location="values")

    parser_mod = ns.parser()
    parser_mod.add_argument("grant", required=True, help="Group grant content")
    parser_mod.add_argument("backend", help="Backend", location="values")

    parser_del = ns.parser()
    parser_del.add_argument("backend", help="Backend", location="values")

    @api.acl_admin_or_moderator_required(message="Not allowed to view groups list")
    @ns.marshal_list_with(group_fields, code=200, description="Success")
    @ns.doc(
        responses={
            403: "Insufficient permissions",
            404: "No backend found",
        },
    )
    def get(self, name=None, backend=None):
        """Returns a list of group

        **GET** method provided by the webservice.

        :returns: Groups
        """
        try:
            handler = getattr(bui, "acl_handler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        ret = []
        for _, loader in handler.backends.items():
            groups = loader.groups
            if groups:
                for _id, group in groups.items():
                    append = {
                        "id": _id.lstrip("@"),
                        "grant": group.get("grants", ""),
                        "members": group.get("members", []),
                        "backend": loader.name,
                    }
                    if name and name == append["id"]:
                        if (
                            backend and backend == append["backend"]
                        ) or backend is None:
                            return [append]
                    ret.append(append)
        return ret

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(message="Not allowed to create groups")
    @ns.expect(parser_add)
    @ns.doc(
        responses={
            200: "Request performed with errors",
            201: "Success",
            403: "Not allowed",
            400: "Missing parameters",
            404: "Backend not found",
            500: "Backend does not support this operation",
        },
    )
    def put(self, backend=None):
        """Create a new group"""
        args = self.parser_add.parse_args()
        backend = backend or args["backend"]

        try:
            handler = getattr(bui, "acl_handler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0 or backend not in handler.backends:
            self.abort(404, "No acl backend found")

        loader = handler.backends[backend]

        if loader.add_group is False:
            self.abort(
                500,
                "The '{}' backend does not support group creation" "".format(backend),
            )

        success, message, code = loader.add_group(args["group"], args["grant"])
        status = 201 if success else 200

        bui.audit.logger.info(
            f'created new group {args["group"]} with grants: {args["grant"]}'
        )
        return [[code, message]], status

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(message="Not allowed to delete this group")
    @ns.expect(parser_del)
    @ns.doc(
        responses={
            200: "Request performed with errors",
            201: "Success",
            403: "Not allowed",
            400: "Missing parameters",
            404: "Backend not found",
            500: "Backend does not support this operation",
        },
    )
    def delete(self, name, backend=None):
        """Delete a group"""
        args = self.parser_del.parse_args()
        backend = backend or args["backend"]

        try:
            handler = getattr(bui, "acl_handler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0 or backend not in handler.backends:
            self.abort(404, "No acl backend found")

        loader = handler.backends[backend]

        if loader.del_group is False:
            self.abort(
                500,
                "The '{}' backend does not support group deletion" "".format(backend),
            )

        success, message, code = loader.del_group(name)
        status = 201 if success else 200

        bui.audit.logger.info(f"removed group {name}")
        return [[code, message]], status

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(message="Not allowed to modify this group")
    @ns.expect(parser_mod)
    @ns.doc(
        responses={
            200: "Request performed with errors",
            201: "Success",
            403: "Not allowed",
            400: "Missing parameters",
            404: "Backend not found",
            500: "Backend does not support this operation",
        },
    )
    def post(self, name, backend=None):
        """Change a group"""
        args = self.parser_mod.parse_args()
        backend = backend or args["backend"]

        try:
            handler = getattr(bui, "acl_handler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0 or backend not in handler.backends:
            self.abort(404, "No acl backend found")

        loader = handler.backends[backend]

        if loader.mod_group is False:
            self.abort(
                500,
                "The '{}' backend does not support group modification"
                "".format(backend),
            )

        success, message, code = loader.mod_group(name, args["grant"])
        status = 201 if success else 200

        bui.audit.logger.info(f'updated group {name} with: {args["grant"]}')
        return [[code, message]], status


@ns.route(
    "/acl/grants",
    "/acl/<backend>/grants",
    "/acl/grants/<name>",
    "/acl/<backend>/grants/<name>",
    endpoint="acl_grants",
)
@ns.doc(
    params={
        "name": "Grant name",
        "backend": "Backend name",
    }
)
class AclGrants(Resource):
    """The :class:`burpui.api.admin.AclGrants` resource allows you to
    retrieve a list of grants and to add/update/delete them if your
    acl backend support those actions.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    parser_add = ns.parser()
    parser_add.add_argument("grant", required=True, help="Grant name")
    parser_add.add_argument("content", required=True, help="Grant content")
    parser_add.add_argument("backend", help="Backend")

    parser_mod = ns.parser()
    parser_mod.add_argument("content", required=True, help="Grant content")
    parser_mod.add_argument("backend", help="Backend")

    parser_del = ns.parser()
    parser_del.add_argument("backend", help="Backend", location="values")

    @api.acl_admin_or_moderator_required(message="Not allowed to view grants list")
    @ns.marshal_list_with(grant_fields, code=200, description="Success")
    @ns.doc(
        responses={
            403: "Insufficient permissions",
            404: "No backend found",
        },
    )
    def get(self, name=None, backend=None):
        """Returns a list of grants

        **GET** method provided by the webservice.

        :returns: Grants
        """
        try:
            handler = getattr(bui, "acl_handler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        ret = []
        for _, loader in handler.backends.items():
            grants = loader.grants
            if grants:
                for _id, grant in grants.items():
                    append = {
                        "id": _id,
                        "grant": json.dumps(grant),
                        "backend": loader.name,
                    }
                    if name and name == _id:
                        if (backend and backend == loader.name) or backend is None:
                            return [append]
                    ret.append(append)
        return ret

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(message="Not allowed to create grants")
    @ns.expect(parser_add)
    @ns.doc(
        responses={
            200: "Request performed with errors",
            201: "Success",
            403: "Not allowed",
            400: "Missing parameters",
            404: "Backend not found",
            500: "Backend does not support this operation",
        },
    )
    def put(self, backend=None):
        """Create a new grant"""
        args = self.parser_add.parse_args()
        backend = backend or args["backend"]

        try:
            handler = getattr(bui, "acl_handler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0 or backend not in handler.backends:
            self.abort(404, "No acl backend found")

        loader = handler.backends[backend]

        if loader.add_grant is False:
            self.abort(
                500,
                "The '{}' backend does not support grant creation" "".format(backend),
            )

        success, message, code = loader.add_grant(args["grant"], args["content"])
        status = 201 if success else 200

        bui.audit.logger.info(f'added grant {args["grant"]} with: {args["content"]}')
        return [[code, message]], status

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(message="Not allowed to delete this grant")
    @ns.expect(parser_del)
    @ns.doc(
        responses={
            200: "Request performed with errors",
            201: "Success",
            403: "Not allowed",
            400: "Missing parameters",
            404: "Backend not found",
            500: "Backend does not support this operation",
        },
    )
    def delete(self, name, backend=None):
        """Delete a grant"""
        args = self.parser_del.parse_args()
        backend = backend or args["backend"]

        try:
            handler = getattr(bui, "acl_handler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0 or backend not in handler.backends:
            self.abort(404, "No acl backend found")

        loader = handler.backends[backend]

        if loader.del_grant is False:
            self.abort(
                500,
                "The '{}' backend does not support grant deletion" "".format(backend),
            )

        success, message, code = loader.del_grant(name)
        status = 201 if success else 200

        bui.audit.logger.info(f"removed grant {name}")
        return [[code, message]], status

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(message="Not allowed to modify this grant")
    @ns.expect(parser_mod)
    @ns.doc(
        responses={
            200: "Request performed with errors",
            201: "Success",
            403: "Not allowed",
            400: "Missing parameters",
            404: "Backend not found",
            500: "Backend does not support this operation",
        },
    )
    def post(self, name, backend=None):
        """Change a grant"""
        args = self.parser_mod.parse_args()
        backend = backend or args["backend"]

        try:
            handler = getattr(bui, "acl_handler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0 or backend not in handler.backends:
            self.abort(404, "No acl backend found")

        loader = handler.backends[backend]

        if loader.mod_grant is False:
            self.abort(
                500,
                "The '{}' backend does not support grant modification"
                "".format(backend),
            )

        success, message, code = loader.mod_grant(name, args["content"])
        status = 201 if success else 200

        bui.audit.logger.info(f'updated grant {name} with: {args["content"]}')
        return [[code, message]], status


@ns.route("/acl/backend/<backend>", endpoint="acl_backend")
class AclBackend(Resource):
    """The :class:`burpui.api.admin.AclBackend` resource allows you to
    retrieve a given ACL backend with its capabilities.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    @api.acl_admin_or_moderator_required(message="Not allowed to view backends list")
    @ns.marshal_with(acl_backend_fields, code=200, description="Success")
    @ns.doc(
        responses={
            403: "Insufficient permissions",
            404: "No backend found",
        },
    )
    def get(self, backend):
        """Returns a given ACL backend

        **GET** method provided by the webservice.

        :returns: Backend
        """
        try:
            handler = getattr(bui, "acl_handler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No authentication backend found")
        if backend not in handler.backends:
            self.abort(404, "ACL backend {} not found".format(backend))
        loader = handler.backends[backend]
        back = {}
        back["name"] = backend
        back["description"] = gettext(loader.__doc__)
        back["type"] = "authorization"
        back["priority"] = getattr(loader, "priority", -1)
        for method in [
            "add_grant",
            "del_grant",
            "mod_grant",
            "add_group",
            "del_group",
            "mod_group",
            "add_group_member",
            "del_group_member",
            "add_moderator",
            "del_moderator",
            "mod_moderator",
            "add_admin",
            "del_admin",
        ]:
            back[method] = getattr(loader, method, False) is not False

        return back


@ns.route("/acl/backends", endpoint="acl_backends")
class AclBackends(Resource):
    """The :class:`burpui.api.admin.AclBackends` resource allows you to
    retrieve a list of ACL backends with their capabilities.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    @api.acl_admin_or_moderator_required(message="Not allowed to view backends list")
    @ns.marshal_list_with(acl_backend_fields, code=200, description="Success")
    @ns.doc(
        responses={
            403: "Insufficient permissions",
            404: "No backend found",
        },
    )
    def get(self):
        """Returns a list of backends

        **GET** method provided by the webservice.

        :returns: Backends
        """
        try:
            handler = getattr(bui, "acl_handler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No authentication backend found")
        ret = []
        for name, backend in handler.backends.items():
            back = {}
            back["name"] = name
            back["description"] = gettext(backend.__doc__)
            back["type"] = "authorization"
            back["priority"] = getattr(backend, "priority", -1)
            for method in [
                "add_grant",
                "del_grant",
                "mod_grant",
                "add_group",
                "del_group",
                "mod_group",
                "add_group_member",
                "del_group_member",
                "add_moderator",
                "del_moderator",
                "mod_moderator",
                "add_admin",
                "del_admin",
            ]:
                back[method] = getattr(backend, method, False) is not False
            ret.append(back)

        return ret


@ns.route(
    "/auth/users",
    "/auth/<backend>/users",
    "/auth/users/<name>",
    "/auth/<backend>/users/<name>",
    endpoint="auth_users",
)
@ns.doc(
    params={
        "name": "Username",
        "backend": "Authentication backend",
    }
)
class AuthUsers(Resource):
    """The :class:`burpui.api.admin.AuthUsers` resource allows you to
    retrieve a list of users and to add/update/delete them if your
    authentication backend support those actions.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    parser_add = ns.parser()
    parser_add.add_argument("username", required=True, help="Username")
    parser_add.add_argument("password", required=True, help="Password")
    parser_add.add_argument("backend", help="Backend")

    parser_mod = ns.parser()
    parser_mod.add_argument("password", required=True, help="Password")
    parser_mod.add_argument("backend", help="Backend")
    parser_mod.add_argument("old_password", required=False, help="Old password")

    parser_del = ns.parser()
    parser_del.add_argument("backend", help="Backend")

    @api.acl_admin_or_moderator_required(message="Not allowed to view users list")
    @ns.marshal_list_with(user_fields, code=200, description="Success")
    @ns.doc(
        responses={
            403: "Insufficient permissions",
            404: "No backend found",
        },
    )
    def get(self, name=None, backend=None):
        """Returns a list of users

        **GET** method provided by the webservice.

        :returns: Users
        """
        try:
            handler = getattr(bui, "uhandler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No authentication backend found")
        ret = []
        for _, backend in handler.backends.items():
            loader = backend.loader
            preload_users = getattr(backend, "preload_users", True)
            try:
                users = getattr(loader, "users")
            except AttributeError:
                continue
            if users:
                if isinstance(users, list):
                    for user in users:
                        append = {
                            "id": backend.user(user).get_id()
                            if preload_users
                            else user,
                            "name": user,
                            "backend": backend.name,
                        }
                        if name and name == append["id"]:
                            if (
                                backend and backend == append["backend"]
                            ) or backend is None:
                                return append
                        ret.append(append)
                elif isinstance(users, dict):
                    for user, _ in users.items():
                        append = {
                            "id": backend.user(user).get_id(),
                            "name": user,
                            "backend": backend.name,
                        }
                        if name and name == append["id"]:
                            if (
                                backend and backend == append["backend"]
                            ) or backend is None:
                                return append
                        ret.append(append)
        return ret

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(message="Not allowed to create users")
    @ns.expect(parser_add)
    @ns.doc(
        responses={
            200: "Request performed with errors",
            201: "Success",
            403: "Not allowed",
            400: "Missing parameters",
            404: "Backend not found",
            500: "Backend does not support this operation",
        },
    )
    def put(self, name=None, backend=None):
        """Create a new user"""
        args = self.parser_add.parse_args()
        username = name or args["username"]
        backend = backend or args["backend"]

        try:
            handler = getattr(bui, "uhandler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0 or backend not in handler.backends:
            self.abort(404, "No authentication backend found")

        backend = handler.backends[backend]

        if backend.add_user is False:
            self.abort(
                500,
                "The '{}' backend does not support user creation" "".format(backend),
            )

        success, message, code = backend.add_user(username, args["password"])
        status = 201 if success else 200

        bui.audit.logger.info(f"created new user: {username}")
        return [[code, message]], status

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(message="Not allowed to delete this user")
    @ns.expect(parser_del)
    @ns.doc(
        responses={
            200: "Request performed with errors",
            201: "Success",
            403: "Not allowed",
            400: "Missing parameters",
            404: "Backend not found",
            500: "Backend does not support this operation",
        },
    )
    def delete(self, name, backend=None):
        """Delete a user"""
        args = self.parser_del.parse_args()
        backend = backend or args["backend"]

        try:
            handler = getattr(bui, "uhandler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0 or backend not in handler.backends:
            self.abort(404, "No authentication backend found")

        backend = handler.backends[backend]

        if backend.del_user is False:
            self.abort(
                500,
                "The '{}' backend does not support user deletion" "".format(backend),
            )

        success, message, code = backend.del_user(name)
        status = 201 if success else 200

        bui.audit.logger.info(f"removed user: {name}")
        return [[code, message]], status

    @api.disabled_on_demo()
    @api.acl_own_or_admin_or_moderator(
        key="name", message="Not allowed to modify this user"
    )
    @ns.expect(parser_mod)
    @ns.doc(
        responses={
            200: "Request performed with errors",
            201: "Success",
            403: "Not allowed",
            400: "Missing parameters",
            404: "Backend not found",
            500: "Backend does not support this operation",
        },
    )
    def post(self, name, backend=None):
        """Change user password"""
        args = self.parser_mod.parse_args()
        backend = backend or args["backend"]
        is_moderator = True

        if not current_user.is_anonymous:
            is_moderator = (
                current_user.acl.is_admin() or current_user.acl.is_moderator()
            )

        if not is_moderator and not args["old_password"]:
            self.abort(400, "Old password required")

        try:
            handler = getattr(bui, "uhandler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0 or backend not in handler.backends:
            self.abort(404, "No authentication backend found")

        backend = handler.backends[backend]

        if backend.change_password is False:
            self.abort(
                500,
                "The '{}' backend does not support user modification"
                "".format(backend),
            )

        success, message, code = backend.change_password(
            name, args["password"], args.get("old_password")
        )
        status = 201 if success else 200

        bui.audit.logger.info(f"changed password of user {name}")
        return [[code, message]], status


@ns.route("/auth/backend/<backend>", endpoint="auth_backend")
class AuthBackend(Resource):
    """The :class:`burpui.api.admin.AuthBackend` resource allows you to
    retrieve a given authentication backend and its capabilities.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    @api.acl_admin_or_moderator_required(message="Not allowed to view backends list")
    @ns.marshal_with(auth_backend_fields, code=200, description="Success")
    @ns.doc(
        responses={
            403: "Insufficient permissions",
            404: "No backend found",
        },
    )
    def get(self, backend):
        """Returns a given authentication backend

        **GET** method provided by the webservice.

        :returns: Backend
        """
        try:
            handler = getattr(bui, "uhandler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No authentication backend found")
        if backend not in handler.backends:
            self.abort(404, "No authentication backend {} found".format(backend))
        back = handler.backends[backend]
        ret = {
            "name": backend,
            "description": gettext(back.__doc__),
            "type": "authentication",
            "priority": getattr(back, "priority", -1),
            "add": getattr(back, "add_user", False) is not False,
            "del": getattr(back, "del_user", False) is not False,
            "mod": getattr(back, "change_password", False) is not False,
        }

        return ret


@ns.route("/auth/backends", endpoint="auth_backends")
class AuthBackends(Resource):
    """The :class:`burpui.api.admin.AuthBackends` resource allows you to
    retrieve a list of backends and to add/update/delete users if your
    authentication backend support those actions.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    @api.acl_admin_or_moderator_required(message="Not allowed to view backends list")
    @ns.marshal_list_with(auth_backend_fields, code=200, description="Success")
    @ns.doc(
        responses={
            403: "Insufficient permissions",
            404: "No backend found",
        },
    )
    def get(self):
        """Returns a list of backends

        **GET** method provided by the webservice.

        :returns: Backends
        """
        try:
            handler = getattr(bui, "uhandler")
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No authentication backend found")
        ret = []
        for name, backend in handler.backends.items():
            ret.append(
                {
                    "name": name,
                    "description": gettext(backend.__doc__),
                    "type": "authentication",
                    "priority": backend.priority,
                    "add": backend.add_user is not False,
                    "del": backend.del_user is not False,
                    "mod": backend.change_password is not False,
                }
            )

        return ret


@ns.route("/session/<user>", "/session/<user>/<uuid:id>", endpoint="other_sessions")
@ns.doc(
    params={
        "user": "User to get sessions from",
        "id": "Session id",
    }
)
class OtherSessions(Resource):
    """The :class:`burpui.api.admin.OtherSessions` resource allows you to
    retrieve a list of sessions for a given user.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    @ns.marshal_list_with(session_fields, code=200, description="Success")
    @ns.doc(
        responses={
            403: "Insufficient permissions",
            404: "User not found",
        },
    )
    def get(self, user=None, id=None):
        """Returns a list of sessions

        **GET** method provided by the webservice.

        :returns: Sessions
        """
        if id:
            return session_manager.get_session_by_id(str(id))
        if not user:
            self.abort(404, "User not found")
        return session_manager.get_user_sessions(user)


@ns.route("/me/session", "/me/session/<uuid:id>", endpoint="user_sessions")
@ns.doc(
    params={
        "id": "Session id",
    }
)
class MySessions(Resource):
    """The :class:`burpui.api.admin.MySessions` resource allows you to
    retrieve a list of sessions and invalidate them for the current user.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    @ns.marshal_list_with(session_fields, code=200, description="Success")
    @ns.doc(
        responses={
            403: "Insufficient permissions",
            404: "User not found",
        },
    )
    def get(self, id=None):
        """Returns a list of sessions

        **GET** method provided by the webservice.

        :returns: Sessions
        """
        if id:
            return session_manager.get_session_by_id(str(id))
        user = getattr(current_user, "name", None)
        if not user:
            self.abort(404, "User not found")
        return session_manager.get_user_sessions(user)

    @api.disabled_on_demo()
    @ns.doc(
        responses={
            201: "Success",
            403: "Insufficient permissions",
            404: "User or session not found",
            400: "Wrong request",
        }
    )
    def delete(self, id=None):
        """Delete a given session

        Note: ``id`` is mandatory
        """
        if not id:
            self.abort(400, "Missing id")
        user = getattr(current_user, "name", None)
        if not user:
            self.abort(404, "User not found")
        store = session_manager.get_session_by_id(str(id))
        if not store:
            self.abort("Session not found")
        if store.user != user:
            if (
                not current_user.is_anonymous
                and not current_user.acl.is_admin()
                and not current_user.acl.is_moderator()
            ):
                self.abort(403, "Insufficient permissions")
            if current_user.acl.is_moderator() and meta_grants.is_admin(store.user):
                self.abort(403, "Insufficient permissions")
        if session_manager.invalidate_session_by_id(store.uuid):
            session_manager.delete_session_by_id(store.uuid)

        bui.audit.logger.info(f"removed session {store.id} of {store.user}")
        return [NOTIF_OK, "Session {} successfully revoked".format(id)], 201



# -*- coding: utf8 -*-
"""
.. module:: burpui.misc.backend.utils.burp2
    :platform: Unix
    :synopsis: Burp-UI burp utils module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import re
import subprocess
import json
import datetime

from select import select

from ...._compat import to_bytes, to_unicode
from ....exceptions import BUIserverException
from ....security import sanitize_string
from ....tools.logging import logger
from .constant import (
    BURP_LIST_BATCH,
    BURP_MINIMAL_VERSION,
    BURP_STATUS_FORMAT_V2,
    BURP_STATUS_DELIMITER,
)


class Monitor(object):
    """The :class:`burpui.misc.backend.utils.burp2.Monitor` class provides a ``burp-2``
    Monitor object to interact with the server.

    :param app: ``Burp-UI`` server instance in order to access logger
                   and/or some global settings
    :type app: :class:`burpui.engines.server.BUIServer`

    :param burpbin: Burp binary path
    :type burpbin: str

    :param burpconf: Burp configuration path
    :type burpconf: str

    :param timeout: Time to wait for an answer
    :type timeout: int
    """

    logger = logger

    # cache status results
    _status_cache = {}
    _last_status_cleanup = datetime.datetime.now()
    _time_to_cache = datetime.timedelta(seconds=3)

    _ignore_logs = re.compile(r"^Server version: (\d+\.\d+\.\d+).*$")

    def __init__(self, burpbin, burpconf, app=None, timeout=5, ident=None):
        """
        :param app: ``Burp-UI`` server instance in order to access logger
                       and/or some global settings
        :type app: :class:`burpui.engines.server.BUIServer`

        :param burpbin: Burp binary path
        :type burpbin: str

        :param burpconf: Burp configuration path
        :type burpconf: str

        :param timeout: Time to wait for an answer
        :type timeout: int
        """
        self.burpbin = burpbin
        self.burpconf = burpconf
        self.timeout = timeout
        self.app = app
        self.proc = None
        self._client_version = None
        self._server_version = None
        self.batch_list_supported = False
        self.status_delimiter = False
        self.ident = ident or id(self)

        self._burp_client_ok = False
        version = ""
        # check the burp version because this backend only supports clients
        # newer than BURP_MINIMAL_VERSION
        try:
            if self.burpbin:
                # the '--version' flag changed in burp 2.2.12
                cmd = [self.burpbin, "-V"]
                version = None
                try:
                    version = to_unicode(
                        subprocess.check_output(cmd, stderr=subprocess.DEVNULL)
                    ).rstrip()
                except subprocess.CalledProcessError:
                    pass
                if version is None:
                    cmd = [self.burpbin, "-v"]
                    version = to_unicode(
                        subprocess.check_output(cmd, stderr=subprocess.DEVNULL)
                    ).rstrip()
                if version < BURP_MINIMAL_VERSION and getattr(self.app, "strict", True):
                    self.logger.critical(
                        f"Your burp version ({version}) does not fit the minimal"
                        f" requirements: {BURP_MINIMAL_VERSION}"
                    )
                elif version >= BURP_MINIMAL_VERSION:
                    self._burp_client_ok = True
        except subprocess.CalledProcessError as exp:
            if getattr(self.app, "strict", True):
                self.logger.critical(f"Unable to determine your burp version: {exp}")

        self._client_version = version.replace("burp-", "")

        if self.app and not self.app.config["BUI_CLI"]:
            try:
                # make the connection
                self._spawn_burp(True)
                self.status()
            except BUIserverException:
                pass
            except OSError as exp:
                msg = str(exp)
                self.logger.critical(msg)

    @property
    def client_version(self):
        return self._client_version or ""

    @property
    def server_version(self):
        if self._server_version is None:
            try:
                self.status()
            except BUIserverException:
                return ""
        return self._server_version or ""

    @property
    def alive(self):
        # clients may be idle for some time, in that case we may need to start them again
        if not self._proc_is_alive():
            try:
                self._spawn_burp()
            except OSError:
                pass
        return self._proc_is_alive()

    def _exit(self):
        """try not to leave child process server side"""
        self.logger.debug(f"Exiting {self.ident}")
        self._terminate_burp()
        self._kill_burp()

    def __exit__(self, typ, value, traceback):
        self._exit()

    def __del__(self):
        self._exit()

    def _kill_burp(self):
        """Terminate the process"""
        if self._proc_is_alive():
            try:
                self.proc.terminate()
            except Exception:
                pass
        if self._proc_is_alive():
            try:
                self.proc.kill()
            except Exception:
                pass

    def _terminate_burp(self):
        """Terminate cleanly the process"""
        if self._proc_is_alive():
            self.proc.stdin.close()
            self.proc.wait()
            self.proc = None

    def _spawn_burp(self, verbose=False):
        """Launch the burp client process"""
        if not self._burp_client_ok:
            raise BUIserverException("No suitable burp client found")
        cmd = [self.burpbin, "-c", self.burpconf, "-a", "m"]
        self.proc = subprocess.Popen(
            cmd,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            shell=False,
            bufsize=0,
        )
        if not self._proc_is_alive():
            details = ""
            if verbose:
                details = ":\n"
                out, _ = self.proc.communicate()
                details += to_unicode(out)
            raise OSError(f"Unable to spawn burp process{details}")
        _, write, _ = select([], [self.proc.stdin], [], self.timeout)
        if self.proc.stdin not in write:
            self._kill_burp()
            raise OSError("Unable to setup burp client")
        self.proc.stdin.write(to_bytes("j:pretty-print-off\n"))
        self._read_proc_stdout(self.timeout)
        # try to switch to new JSON status
        if self.client_version < BURP_STATUS_FORMAT_V2:
            self.proc.stdin.write(to_bytes("j:peer_version=2.1.10\n"))
            self._read_proc_stdout(self.timeout)
        if self.status_delimiter:
            self.proc.stdin.write(to_bytes("j:response-markers-on\n"))
            self._read_proc_stdout(self.timeout, "j:response-markers-on")

    def _proc_is_alive(self):
        """Check if the burp client process is still alive"""
        if self.proc:
            return self.proc.poll() is None
        return False

    def _is_ignored(self, jso, watching=None):
        """We ignore the 'logline' lines"""
        if not jso:
            return True
        if not self._server_version:
            if "logline" in jso:
                ret = self._ignore_logs.search(jso["logline"])
                if ret:
                    self._server_version = ret.group(1)
                    if self.server_version >= BURP_LIST_BATCH:
                        self.batch_list_supported = True
                    if self.server_version >= BURP_STATUS_DELIMITER:
                        self.status_delimiter = True
        if (
            self.status_delimiter
            and watching is None
            and ("response-start" in jso or "response-end" in jso)
        ):
            return True
        return "logline" in jso

    @staticmethod
    def _is_warning(jso):
        """Returns True if the document is a warning"""
        if not jso:
            return False
        return "warning" in jso

    @staticmethod
    def _is_valid_json(doc):
        """Determine if the retrieved string is a valid json document or not"""
        try:
            jso = json.loads(doc)
            return jso
        except ValueError:
            return None

    def _read_proc_stdout(self, timeout, watching=None):
        """reads the burp process stdout and returns a document or None"""
        doc = ""
        tmp = ""
        jso = None
        cache = {}
        if watching:
            watching = watching.rstrip()
        while True:
            try:
                if not self._proc_is_alive():
                    raise OSError("process died while reading its output")
                read, _, _ = select([self.proc.stdout], [], [], timeout)
                if self.proc.stdout not in read:
                    raise TimeoutError("Read operation timed out")
                tmp += to_unicode(self.proc.stdout.readline()).rstrip("\n")
                jso = self._is_valid_json(tmp)
                # if the string is a valid json and looks like a logline, we
                # simply ignore it
                if jso and self._is_ignored(jso, watching):
                    tmp = ""
                    continue
                elif jso:
                    if not self.status_delimiter or (
                        self.status_delimiter and watching is None
                    ):
                        doc = tmp
                        break
                    start = jso.get("response-start")
                    end = jso.get("response-end")
                    if not start and not end:
                        cache["raw"] = tmp
                        cache["json"] = jso
                    elif start and start != watching:
                        doc = ""
                        jso = None
                        break
                    elif end and end != watching:
                        doc = ""
                        jso = None
                        break
                    elif end:
                        doc = cache.get("raw", "")
                        jso = cache.get("json")
                        break
                    tmp = ""
            except (TimeoutError, IOError, OSError) as exp:
                # the os throws an exception if there is no data or timeout
                self.logger.warning(str(exp))
                self._kill_burp()
                return None, None
        return jso, doc

    def _cleanup_cache(self):
        now = datetime.datetime.now()
        if now - self._last_status_cleanup > self._time_to_cache:
            self._status_cache.clear()
            self._last_status_cleanup = now

    def status(self, query="c:\n", timeout=None, cache=True, raw=False):
        """Send the given query to the status port and parses the output"""
        try:
            timeout = timeout or self.timeout
            query = sanitize_string(query.rstrip())
            self.logger.info(
                f"{self.ident} - query: '{query}' (cache: {cache}, raw: {raw})"
            )
            query = "{0}\n".format(query)

            self._cleanup_cache()
            # return cached results
            key = f"{query}-{raw}"
            if cache and key in self._status_cache:
                return self._status_cache[key]

            if not self._proc_is_alive():
                self._spawn_burp()

            _, write, _ = select([], [self.proc.stdin], [], timeout)
            if self.proc.stdin not in write:
                raise TimeoutError("Write operation timed out")
            self.proc.stdin.write(to_bytes(query))
            jso, doc = self._read_proc_stdout(timeout, query)
            if self._is_warning(jso):
                self.logger.warning(jso["warning"])
                self.logger.debug("Nothing interesting to return")
                return None

            if raw:
                res = doc
            else:
                res = jso

            self.logger.debug(f"{self.ident} => {res}")

            if cache:
                self._status_cache[key] = res

            return res
        except TimeoutError as exp:
            msg = f"Cannot send command: {exp}"
            self.logger.error(msg)
            self._kill_burp()
            if getattr(self.app, "strict", True):
                raise BUIserverException(msg)
        except (OSError, Exception) as exp:
            msg = f"Cannot launch burp process: {exp}"
            self.logger.error(msg)
            if getattr(self.app, "strict", True):
                raise BUIserverException(msg)
        return None



"""initial

Revision ID: 225d9b2f0fb1
Revises: None
Create Date: 2016-07-18 11:22:27.613921

"""

# revision identifiers, used by Alembic.
revision = "225d9b2f0fb1"
down_revision = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "task",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("uuid", sa.String(length=256), nullable=True),
        sa.Column("task", sa.String(length=256), nullable=True),
        sa.Column("user", sa.String(length=256), nullable=True),
        sa.Column("timestamp", sa.DateTime(), nullable=True),
        sa.Column("expire", sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("uuid"),
    )
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("task")
    ### end Alembic commands ###



# -*- coding: utf8 -*-
"""
.. module:: burpui.models
    :platform: Unix
    :synopsis: Burp-UI DB models module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import datetime

from .ext.sql import db
from flask import current_app, session
from .engines.server import BUIServer  # noqa

app = current_app  # type: BUIServer


class Task(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    uuid = db.Column(db.String(256), unique=True)
    task = db.Column(db.String(256))
    user = db.Column(db.String(256), nullable=True)
    timestamp = db.Column(db.DateTime, default=datetime.datetime.utcnow)
    expire = db.Column(db.DateTime, nullable=True)

    def __init__(self, uuid, task, user=None, expire=None):
        self.task = task
        self.uuid = uuid
        self.user = user
        if expire is not None:
            self.expire = datetime.datetime.utcnow() + expire

    def __repr__(self):
        return "<Task {}-{} ({}, {}/{})>".format(
            self.task, self.uuid, self.user, self.timestamp, self.expire
        )


class Session(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    uuid = db.Column(db.String(256), unique=True)
    user = db.Column(db.String(256))
    ip = db.Column(db.String(256), nullable=True)
    ua = db.Column(db.String(2048), nullable=True)
    timestamp = db.Column(db.DateTime, default=datetime.datetime.utcnow)
    expire = db.Column(db.DateTime, nullable=True)
    permanent = db.Column(db.Boolean)
    api = db.Column(db.Boolean)

    def __init__(self, uuid, user, ip=None, ua=None, permanent=False, api=False):
        self.uuid = uuid
        self.user = user
        self.ip = ip
        self.ua = ua
        self.permanent = permanent
        self.api = api
        if self.permanent:
            self.expire = datetime.datetime.utcnow() + app.permanent_session_lifetime

    def refresh(self, ip=None):
        self.timestamp = datetime.datetime.utcnow()
        if ip:
            self.ip = ip
        if self.permanent:
            self.expire = datetime.datetime.utcnow() + app.permanent_session_lifetime
            if "remember" not in session:
                session["remember"] = "set"
        db.session.commit()

    def __repr__(self):
        return "<Session {} ({}, {}, {})>".format(
            self.uuid, self.user, self.ip, self.ua
        )


class Pref(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    user = db.Column(db.String(256))
    key = db.Column(db.String(256), nullable=False)
    value = db.Column(db.Text)

    def __init__(self, user, key, value=None):
        self.user = user
        self.key = key
        self.value = value

    def __repr__(self):
        return "<Pref {}: {} = {}>".format(self.user, self.key, self.value)


class Hidden(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    user = db.Column(db.String(256), nullable=False)
    client = db.Column(db.String(4096))
    server = db.Column(db.String(4096))

    def __init__(self, user, client=None, server=None):
        self.user = user
        self.client = client
        self.server = server

    def __repr__(self):
        return "<Hidden {}: {} ({})".format(self.user, self.client, self.server)


def lazy_loading():
    """This function is used so the above classes/tables are detected by alembic.
    It does nothing except allowing us to compile this module"""
    pass



# -*- coding: utf8 -*-
"""
.. module:: burpui.api
    :platform: Unix
    :synopsis: Burp-UI api module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>


"""
import os
import sys
import uuid
import hashlib

from flask import Blueprint, Response, request, current_app, session, abort, g
from flask_restx import Api as ApiPlus
from flask_login import current_user
from importlib import import_module
from functools import wraps

from .._compat import to_bytes
from ..desc import __version__, __release__, __url__, __doc__
from ..engines.server import BUIServer  # noqa
from ..exceptions import BUIserverException
from ..config import config
from ..tools.logging import logger

bui = current_app  # type: BUIServer
EXEMPT_METHODS = set(["OPTIONS"])


def force_refresh():
    return request.headers.get("X-No-Cache", False) is not False or getattr(
        g, "DONOTCACHE", False
    )


def cache_key():
    key = "{}-{}-{}-{}-{}-{}".format(
        session.get("login", uuid.uuid4()),
        request.path,
        request.values,
        request.headers.get("X-Session-Tag", ""),
        request.cookies,
        session.get("language", ""),
    )
    key = hashlib.sha256(to_bytes(key)).hexdigest()
    return key


def api_login_required(func):
    """Custom login decorator that is able to parse Basic credentials as well as
    Cookies set with the traditional login.
    """

    @wraps(func)
    def decorated_view(*args, **kwargs):
        """decorator"""
        if request.method in EXEMPT_METHODS:  # pragma: no cover
            return func(*args, **kwargs)
        # 'func' is a Flask.view.MethodView so we have access to some special
        # params
        cls = func.view_class
        login_required = getattr(cls, "login_required", True)
        if (
            bui.auth != "none"
            and login_required
            and not bui.config.get("LOGIN_DISABLED", False)
        ):
            if not current_user.is_authenticated:
                if request.headers.get("X-From-UI", False):
                    abort(403)
                return Response(
                    "Could not verify your access level for that URL.\n"
                    "You have to login with proper credentials",
                    401,
                    {"WWW-Authenticate": 'Basic realm="Login Required"'},
                )
        return func(*args, **kwargs)

    return decorated_view


def check_acl(func):
    """Custom decorator to check if the ACL are in use or not"""

    @wraps(func)
    def decorated_view(*args, **kwargs):
        if request.method in EXEMPT_METHODS:  # pragma: no cover
            return func(*args, **kwargs)
        # 'func' is a Flask.view.MethodView so we have access to some special
        # params
        cls = func.view_class
        login_required = getattr(cls, "login_required", True)
        if (
            bui.auth != "none"
            and login_required
            and not bui.config.get("LOGIN_DISABLED", False)
        ):
            if current_user.is_anonymous:
                abort(403)
        return func(*args, **kwargs)

    return decorated_view


class Api(ApiPlus):
    """Wrapper class around :class:`flask_restx.Api`"""

    logger = logger
    # TODO: should use global object instead of reference
    loaded = False
    release = __release__
    __doc__ = __doc__
    __url__ = __url__
    CELERY_REQUIRED = ["tasks"]

    def load_all(self):
        if config["WITH_LIMIT"]:
            try:
                from ..ext.limit import limiter

                self.decorators.append(limiter.limit(config["BUI_RATIO"]))
            except ImportError:
                self.logger.warning("Unable to import limiter module")
        """hack to automatically import api modules"""
        if not self.loaded:
            sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))
            self.loaded = True
            for f in os.listdir(__path__[0]):
                name, ext = os.path.splitext(f)
                if (
                    os.path.isfile(os.path.join(__path__[0], f))
                    and ext == ".py"
                    and name not in ["__init__", ".", ".."]
                ):
                    mod = "." + name
                    if name not in self.CELERY_REQUIRED or config["WITH_CELERY"]:
                        self.logger.debug("Loading API module: {}".format(mod))
                        try:
                            import_module(mod, __name__)
                        except Exception as exc:  # pragma: no cover
                            self.logger.critical(
                                "Unable to load {}:\n{}".format(
                                    mod,
                                    exc,
                                ),
                                exc_info=exc,
                                stack_info=True,
                            )
                    else:
                        self.logger.warning("Skipping API module: {}".format(mod))

    def acl_admin_required(self, message="Access denied", code=403):
        def decorator(func):
            @wraps(func)
            def decorated(resource, *args, **kwargs):
                if not current_user.is_anonymous and not current_user.acl.is_admin():
                    resource.abort(code, message)
                return func(resource, *args, **kwargs)

            return decorated

        return decorator

    def acl_admin_or_moderator_required(self, message="Access denied", code=403):
        def decorator(func):
            @wraps(func)
            def decorated(resource, *args, **kwargs):
                if (
                    not current_user.is_anonymous
                    and not current_user.acl.is_admin()
                    and not current_user.acl.is_moderator()
                ):
                    resource.abort(code, message)
                return func(resource, *args, **kwargs)

            return decorated

        return decorator

    def acl_own_or_admin(self, key="name", message="Access denied", code=403):
        def decorator(func):
            @wraps(func)
            def decorated(resource, *args, **kwargs):
                if key not in kwargs:  # pragma: no cover
                    resource.abort(500, "key '{}' not found".format(key))
                if (
                    kwargs[key] != current_user.name
                    and not current_user.is_anonymous
                    and not current_user.acl.is_admin()
                ):
                    resource.abort(code, message)
                return func(resource, *args, **kwargs)

            return decorated

        return decorator

    def acl_own_or_admin_or_moderator(
        self, key="name", message="Access denied", code=403
    ):
        def decorator(func):
            @wraps(func)
            def decorated(resource, *args, **kwargs):
                if key not in kwargs:  # pragma: no cover
                    resource.abort(500, "key '{}' not found".format(key))
                if (
                    kwargs[key] != current_user.name
                    and not current_user.is_anonymous
                    and not current_user.acl.is_admin()
                    and not current_user.acl.is_moderator()
                ):
                    resource.abort(code, message)
                return func(resource, *args, **kwargs)

            return decorated

        return decorator

    def disabled_on_demo(self):
        def decorator(func):
            @wraps(func)
            def decorated(resource, *args, **kwargs):
                if config["BUI_DEMO"]:
                    resource.abort(
                        405, "Sorry, this feature is not available on the demo"
                    )
                return func(resource, *args, **kwargs)

            return decorated

        return decorator


apibp = Blueprint("api", __name__, url_prefix="/api")
api = Api(
    apibp,
    title="Burp-UI API",
    description="Burp-UI API to interact with burp",
    version=__version__,
    doc="/doc",
    decorators=[api_login_required],
)


@api.errorhandler(BUIserverException)
def handle_bui_server_exception(error):
    """Forward a BUIserverException to the final user

    :param error: Custom exception
    :type error: :class:`burpui.exceptions.BUIserverException`
    """
    bui.logger.error(error)
    return {"message": error.description}, error.code



# -*- coding: utf8 -*-
"""
.. module:: burpui.api.misc
    :platform: Unix
    :synopsis: Burp-UI misc api module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from . import api, cache_key, force_refresh
from ..engines.server import BUIServer  # noqa
from .custom import fields, Resource
from .client import ClientLabels
from ..filter import mask
from ..exceptions import BUIserverException
from ..decorators import browser_cache
from ..ext.cache import cache
from ..ext.i18n import LANGUAGES

from flask import flash, get_flashed_messages, url_for, current_app, session
from flask_login import current_user

import random
import re

bui = current_app  # type: BUIServer
ns = api.namespace("misc", "Misc methods")


def clear_cache(pattern=None):
    """Clear the cache, you can also provide a pattern to only clean matching keys"""
    if pattern is None:
        cache.clear()
    else:
        if hasattr(cache.cache, "_client") and hasattr(cache.cache._client, "keys"):
            if hasattr(cache.cache, "key_prefix") and cache.cache.key_prefix:
                pattern = cache.cache.key_prefix + pattern
            keys = cache.cache._client.keys(pattern)
            cache.cache._client.delete(keys)


counters_fields = ns.model(
    "Counters",
    {
        "phase": fields.String(description="Backup phase"),
        "Total": fields.List(
            fields.Integer,
            description="new/deleted/scanned/unchanged/total",
            attribute="total",
        ),
        "Files": fields.List(
            fields.Integer,
            description="new/deleted/scanned/unchanged/total",
            attribute="files",
        ),
        "Files (encrypted)": fields.List(
            fields.Integer,
            description="new/deleted/scanned/unchanged/total",
            attribute="files_encrypted",
        ),
        "Meta data": fields.List(
            fields.Integer,
            description="new/deleted/scanned/unchanged/total",
            attribute="meta_data",
        ),
        "Meta data (enc)": fields.List(
            fields.Integer,
            description="new/deleted/scanned/unchanged/total",
            attribute="meta_data_encrypted",
        ),
        "Directories": fields.List(
            fields.Integer,
            description="new/deleted/scanned/unchanged/total",
            attribute="directories",
        ),
        "Soft links": fields.List(
            fields.Integer,
            description="new/deleted/scanned/unchanged/total",
            attribute="soft_links",
        ),
        "Hard links": fields.List(
            fields.Integer,
            description="new/deleted/scanned/unchanged/total",
            attribute="hard_links",
        ),
        "Special files": fields.List(
            fields.Integer,
            description="new/deleted/scanned/unchanged/total",
            attribute="special_files",
        ),
        "VSS headers": fields.List(
            fields.Integer,
            description="new/deleted/scanned/unchanged/total",
            attribute="vss_headers",
        ),
        "VSS headers (enc)": fields.List(
            fields.Integer,
            description="new/deleted/scanned/unchanged/total",
            attribute="vss_headers_encrypted",
        ),
        "VSS footers": fields.List(
            fields.Integer,
            description="new/deleted/scanned/unchanged/total",
            attribute="vss_footers",
        ),
        "VSS footers (enc)": fields.List(
            fields.Integer,
            description="new/deleted/scanned/unchanged/total",
            attribute="vss_footers_encrypted",
        ),
        "Grand total": fields.List(
            fields.Integer,
            description="new/deleted/scanned/unchanged/total",
            attribute="grand_total",
        ),
        "warning": fields.Integer(description="Number of warnings so far"),
        "estimated_bytes": fields.Integer(description="Estimated Bytes in backup"),
        "bytes": fields.Integer(description="Bytes in backup"),
        "bytes_in": fields.Integer(description="Bytes received since backup started"),
        "bytes_out": fields.Integer(description="Bytes sent since backup started"),
        "start": fields.String(description="Timestamp of the start date of the backup"),
        "speed": fields.Integer(description="Backup speed", default=-1),
        "timeleft": fields.Integer(description="Estimated time left"),
        "percent": fields.Integer(required=True, description="Percentage done"),
        "path": fields.String(description="File that is currently treated by burp"),
    },
)


@ns.route(
    "/counters",
    "/<server>/counters",
    "/counters/<client>",
    "/<server>/counters/<client>",
    endpoint="counters",
)
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
        "client": "Client name",
    },
)
class Counters(Resource):
    """The :class:`burpui.api.misc.Counters` resource allows you to
    render the *live view* template of a given client.

    This resource is part of the :mod:`burpui.api.api` module.

    An optional ``GET`` parameter called ``serverName`` is supported when running
    in multi-agent mode.
    A mandatory ``GET`` parameter called ``clientName`` is used to know what client we
    are working on.
    """

    parser = ns.parser()
    parser.add_argument(
        "serverName", help="Which server to collect data from when in multi-agent mode"
    )
    parser.add_argument("clientName", help="Client name")
    monitor_fields = ns.model(
        "Monitor",
        {
            "client": fields.String(required=True, description="Client name"),
            "agent": fields.String(description="Server (agent) name"),
            "counters": fields.Nested(
                counters_fields,
                description="Various statistics about the running backup",
            ),
            "labels": fields.List(fields.String, description="List of labels"),
        },
    )

    @ns.marshal_with(monitor_fields, code=200, description="Success")
    @ns.expect(parser)
    @ns.doc(
        responses={
            400: "Missing argument",
            403: "Insufficient permissions",
            404: "Client not found in the running clients list",
        },
    )
    def get(self, server=None, client=None):
        """Returns counters for a given client

        **GET** method provided by the webservice.

        :param name: the client name if any. You can also use the GET parameter
                     'name' to achieve the same thing

        :returns: Counters
        """
        args = self.parser.parse_args()
        server = server or args["serverName"]
        client = client or args["clientName"]
        # Check params
        if not client:
            self.abort(400, "No client name provided")
        # Manage ACL
        if (
            not current_user.is_anonymous
            and not current_user.acl.is_admin()
            and not current_user.acl.is_client_allowed(client, server)
        ):
            self.abort(403, "Not allowed to view '{}' counters".format(client))
        running = bui.client.is_one_backup_running()
        if isinstance(running, dict):
            if server and client not in running[server]:
                self.abort(
                    404,
                    "'{}' not found in the list of running clients for '{}'".format(
                        client, server
                    ),
                )
            else:
                found = False
                for (_, cls) in running.items():
                    if client in cls:
                        found = True
                        break
                if not found:
                    api.bort(404, "'{}' not found in running clients".format(client))
        else:
            if client not in running:
                self.abort(404, "'{}' not found in running clients".format(client))
        try:
            counters = bui.client.get_counters(client, agent=server)
        except BUIserverException:
            counters = {}
        res = {}
        res["client"] = client
        res["agent"] = server
        res["counters"] = counters
        try:
            res["labels"] = ClientLabels._get_labels(client, server)
        except BUIserverException as exp:
            self.abort(500, str(exp))
        return res


@ns.route("/monitor", "/<server>/monitor", endpoint="live")
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
    },
)
class Live(Resource):
    """The :class:`burpui.api.misc.Live` resource allows you to
    retrieve a list of servers that are currently *alive*.

    This resource is part of the :mod:`burpui.api.misc` module.

    An optional ``GET`` parameter called ``serverName`` is supported when running
    in multi-agent mode.
    """

    parser = ns.parser()
    parser.add_argument(
        "serverName", help="Which server to collect data from when in multi-agent mode"
    )
    live_fields = ns.model(
        "Live",
        {
            "client": fields.String(required=True, description="Client name"),
            "agent": fields.String(description="Server (agent) name"),
            "counters": fields.Nested(
                counters_fields,
                description="Various statistics about the running backup",
            ),
            "labels": fields.List(fields.String, description="List of labels"),
        },
    )

    @ns.marshal_list_with(live_fields, code=200, description="Success")
    @ns.expect(parser)
    def get(self, server=None):
        """Returns a list of clients that are currently running a backup

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            [
              {
                'client': 'client1',
                'agent': 'burp1',
                'counters': {
                    'phase': 2,
                    'path': '/etc/some/configuration',
                    '...': '...'
                },
                'labels': [
                    '...'
                ]
              },
              {
                'client': 'client12',
                'agent': 'burp2',
                'counters': {
                    'phase': 3,
                    'path': '/etc/some/other/configuration',
                    '...': '...'
                },
                'labels': [
                    '...'
                ]
              }
            ]


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent mode
        :type server: str

        :returns: The *JSON* described above
        """

        args = self.parser.parse_args()
        server = server or args["serverName"]
        res = []
        is_admin = True
        has_acl = not current_user.is_anonymous

        if has_acl:
            is_admin = current_user.acl.is_admin()

        # ACL
        if (
            has_acl
            and not is_admin
            and server
            and not current_user.acl.is_server_allowed(server)
        ):
            self.abort(403, "You are not allowed to view stats of this server")
        if server:
            running = bui.client.is_one_backup_running(server)
            # ACL
            if mask.has_filters(current_user):
                running = [
                    x
                    for x in running
                    if mask.is_client_allowed(current_user, x, server)
                ]
        else:
            running = bui.client.is_one_backup_running()
        if isinstance(running, dict):
            for (serv, clients) in running.items():
                for client in clients:
                    # ACL
                    if mask.has_filters(current_user) and not mask.is_client_allowed(
                        current_user, client, serv
                    ):
                        continue
                    data = {}
                    data["client"] = client
                    data["agent"] = serv
                    try:
                        data["counters"] = bui.client.get_counters(client, agent=serv)
                    except BUIserverException:
                        data["counters"] = {}
                    try:
                        data["labels"] = ClientLabels._get_labels(client, serv)
                    except BUIserverException:
                        data["labels"] = []
                    res.append(data)
        else:
            for client in running:
                # ACL
                if mask.has_filters(current_user) and not mask.is_client_allowed(
                    current_user, client, server
                ):
                    continue
                data = {}
                data["client"] = client
                try:
                    data["counters"] = bui.client.get_counters(client, agent=server)
                except BUIserverException:
                    data["counters"] = {}
                try:
                    data["labels"] = ClientLabels._get_labels(client)
                except BUIserverException:
                    data["labels"] = []
                res.append(data)
        return res


@ns.route("/alert", endpoint="alert")
class Alert(Resource):
    """The :class:`burpui.api.misc.Alert` resource allows you to propagate a
    message to the next screen.

    This resource is part of the :mod:`burpui.api.misc` module.
    """

    parser = ns.parser()
    parser.add_argument("message", required=True, help="Message to display")
    parser.add_argument(
        "level",
        help="Alert level",
        choices=("danger", "warning", "info", "success", "0", "1", "2", "3"),
        default="danger",
    )

    @ns.expect(parser)
    @ns.doc(
        responses={
            201: "Success",
        },
    )
    def post(self):
        """Propagate a message to the next screen (or whatever reads the session)"""

        def translate(level):
            levels = ["danger", "warning", "info", "success"]
            convert = {"0": "success", "1": "warning", "2": "error", "3": "info"}
            if not level:
                return "danger"
            # return the converted value or the one we already had
            new = convert.get(level, level)
            # if the level is not handled, assume 'danger'
            if new not in levels:
                return "danger"
            return new

        # retrieve last flashed messages so we don't loose anything
        for level, message in get_flashed_messages(with_categories=True):
            flash(message, level)

        args = self.parser.parse_args()
        message = args["message"]
        level = translate(args["level"])
        flash(message, level)
        return {"message": message, "level": level}, 201


@ns.route("/languages", endpoint="languages")
class Languages(Resource):
    """The :class:`burpui.api.misc.Languages` resource allows you to retrieve
    a list of supported languages.

    This resource is part of the :mod:`burpui.api.misc` module.
    """

    wild = fields.Wildcard(fields.String, description="Supported languages")
    languages_fields = ns.model(
        "Languages",
        {
            "*": wild,
        },
    )

    @cache.cached(timeout=3600, key_prefix=cache_key, unless=force_refresh)
    @ns.marshal_with(languages_fields, code=200, description="Success")
    @browser_cache(3600)
    def get(self):
        """Returns a list of supported languages

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            {
              "en": "English",
              "fr": "Français"
            }


        :returns: The *JSON* described above.
        """
        return LANGUAGES


@ns.route("/about", "/<server>/about", endpoint="about")
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
    },
)
class About(Resource):
    """The :class:`burpui.api.misc.About` resource allows you to retrieve
    various informations about ``Burp-UI``

    An optional ``GET`` parameter called ``serverName`` is supported when running
    in multi-agent mode.
    """

    # Login not required on this view
    login_required = False

    parser = ns.parser()
    parser.add_argument(
        "serverName", help="Which server to collect data from when in multi-agent mode"
    )
    burp_fields = ns.model(
        "Burp",
        {
            "name": fields.String(
                required=True, description="Instance name", default="Burp"
            ),
            "client": fields.String(description="Burp client version"),
            "server": fields.String(description="Burp server version"),
        },
    )
    about_fields = ns.model(
        "About",
        {
            "version": fields.String(required=True, description="Burp-UI version"),
            "release": fields.String(description="Burp-UI release (commit number)"),
            "api": fields.String(description="Burp-UI API documentation URL"),
            "burp": fields.Nested(
                burp_fields, as_list=True, description="Burp version"
            ),
        },
    )

    @cache.cached(timeout=3600, key_prefix=cache_key, unless=force_refresh)
    @ns.marshal_with(about_fields, code=200, description="Success")
    @ns.expect(parser)
    @browser_cache(3600)
    def get(self, server=None):
        """Returns various informations about Burp-UI"""
        args = self.parser.parse_args()
        res = {}
        server = server or args["serverName"]
        res["version"] = api.version
        res["release"] = api.release
        res["api"] = url_for("api.doc")
        res["burp"] = []
        cli = bui.client.get_client_version(server)
        srv = bui.client.get_server_version(server)
        multi = {}
        if isinstance(cli, dict):
            for (name, val) in cli.items():
                multi[name] = {"client": val}
        if isinstance(srv, dict):
            for (name, val) in srv.items():
                multi[name]["server"] = val
        if not multi:
            res["burp"].append({"client": cli, "server": srv})
        else:
            for (name, val) in multi.items():
                tmp = val
                tmp.update({"name": name})
                res["burp"].append(tmp)
        return res


@ns.route("/ping", endpoint="ping")
class Ping(Resource):
    """The :class:`burpui.api.misc.Ping` resource allows you to ping the API.
    It is actually a Dummy endpoint that does nothing"""

    # Login not required on this view
    login_required = False

    ping_fields = ns.model(
        "Ping",
        {
            "alive": fields.Boolean(required=True, description="API alive?"),
        },
    )

    @ns.marshal_list_with(ping_fields, code=200, description="Success")
    @ns.doc(
        responses={
            200: "Success",
            403: "Insufficient permissions",
        },
    )
    def get(self):
        """Tells if the API is alive"""
        return {"alive": True}


@ns.route(
    "/history",
    "/history/<client>",
    "/<server>/history",
    "/<server>/history/<client>",
    endpoint="history",
)
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
        "client": "Client name",
    },
)
class History(Resource):
    """The :class:`burpui.api.misc.History` resource allows you to retrieve
    an history of the backups

    An optional ``GET`` parameter called ``serverName`` is supported when
    running in multi-agent mode and ``clientName`` is also allowed to filter
    by client.

    ::

        $('#calendar').fullCalendar({

            eventSources: [

                // your event source
                {
                    events: [ // put the array in the `events` property
                        {
                            title  : 'event1',
                            start  : '2010-01-01'
                        },
                        {
                            title  : 'event2',
                            start  : '2010-01-05',
                            end    : '2010-01-07'
                        },
                        {
                            title  : 'event3',
                            start  : '2010-01-09T12:30:00',
                        }
                    ],
                    color: 'black',     // an option!
                    textColor: 'yellow' // an option!
                }

                // any other event sources...

            ]

        });

    """

    parser = ns.parser()
    parser.add_argument(
        "serverName", help="Which server to collect data from when in multi-agent mode"
    )
    parser.add_argument("clientName", help="Which client to collect data from")
    parser.add_argument("start", help="Return events after this date")
    parser.add_argument("end", help="Return events before this date")

    event_fields = ns.model(
        "Event",
        {
            "title": fields.String(required=True, description="Event name"),
            "start": fields.DateTime(
                dt_format="iso8601",
                description="Start time of the event",
                attribute="date",
            ),
            "end": fields.DateTime(
                dt_format="iso8601", description="End time of the event"
            ),
            "name": fields.String(description="Client name"),
            "backup": fields.BackupNumber(
                description="Backup number", attribute="number"
            ),
            "url": fields.String(description="Callback URL"),
        },
    )
    history_fields = ns.model(
        "History",
        {
            "events": fields.Nested(
                event_fields, as_list=True, description="Events list"
            ),
            "color": fields.String(description="Background color"),
            "textColor": fields.String(description="Text color"),
            "name": fields.String(description="Feed name"),
        },
    )

    @cache.cached(timeout=1800, key_prefix=cache_key, unless=force_refresh)
    @ns.marshal_list_with(history_fields, code=200, description="Success")
    @ns.expect(parser)
    @ns.doc(
        responses={
            200: "Success",
            403: "Insufficient permissions",
        },
    )
    @browser_cache(1800)
    def get(self, client=None, server=None):
        """Returns a list of calendars describing the backups that have been
        completed so far

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            [
              {
                "color": "#7C6F44",
                "events": [
                  {
                    "backup": "0000001",
                    "end": "2015-01-25 13:32:04+01:00",
                    "name": "toto-test",
                    "start": "2015-01-25 13:32:00+01:00",
                    "title": "Client: toto-test, Backup n°0000001",
                    "url": "/client/toto-test"
                  }
                ],
                "name": "toto-test",
                "textColor": "white"
              }
            ]


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent mode
        :type server: str
        :param client: Which client to collect data from
        :type client: str

        :returns: The *JSON* described above
        """
        self._check_acl(client, server)
        return self._get_backup_history(client, server)

    def _check_acl(self, client=None, server=None):
        args = self.parser.parse_args()
        client = client or args["clientName"]
        server = server or args["serverName"]

        if (
            server
            and mask.has_filters(current_user)
            and not mask.is_server_allowed(current_user, server)
        ):
            self.abort(403, "You are not allowed to view this server infos")

        if (
            client
            and mask.has_filters(current_user)
            and not mask.is_client_allowed(current_user, client, server)
        ):
            self.abort(403, "You are not allowed to view this client infos")

    def _get_backup_history(self, client=None, server=None, data=None):
        import arrow

        ret = []
        args = self.parser.parse_args()
        client = client or args["clientName"]
        server = server or args["serverName"]
        moments = {"start": None, "end": None}
        has_filters = mask.has_filters(current_user)

        for moment in moments.keys():
            if moment in args:
                try:
                    if args[moment] is not None:
                        moments[moment] = arrow.get(args[moment]).int_timestamp
                except arrow.parser.ParserError:
                    pass

        if client:
            (color, text) = self.gen_colors(client, server)
            feed = {
                "color": color,
                "textColor": text,
                "events": self.gen_events(client, moments, server, data),
            }
            name = client
            if server:
                name += " on {}".format(server)
            feed["name"] = name
            ret.append(feed)
            return ret
        elif server:
            if data and server in data:
                clients = [{"name": x} for x in data[server].keys()]
            else:
                clients = bui.client.get_all_clients(agent=server, last_attempt=False)
            # manage ACL
            if has_filters:
                clients = [
                    x
                    for x in clients
                    if mask.is_client_allowed(current_user, x["name"], server)
                ]
            for cl in clients:
                (color, text) = self.gen_colors(cl["name"], server)
                feed = {
                    "events": self.gen_events(cl["name"], moments, server, data),
                    "textColor": text,
                    "color": color,
                    "name": "{} on {}".format(cl["name"], server),
                }
                ret.append(feed)
            return ret

        if bui.config["STANDALONE"]:
            if data:
                clients_list = data.keys()
            else:
                try:
                    clients_list = [
                        x["name"]
                        for x in bui.client.get_all_clients(last_attempt=False)
                    ]
                except BUIserverException:
                    clients_list = []
                if has_filters:
                    clients_list = [
                        x
                        for x in clients_list
                        if mask.is_client_allowed(current_user, x)
                    ]
            for cl in clients_list:
                (color, text) = self.gen_colors(cl)
                feed = {
                    "events": self.gen_events(cl, moments, data=data),
                    "textColor": text,
                    "color": color,
                    "name": cl,
                }
                ret.append(feed)
            return ret
        else:
            grants = {}
            for serv in bui.client.servers:
                if has_filters:
                    try:
                        all_clients = [
                            x["name"]
                            for x in bui.client.get_all_clients(
                                serv, last_attempt=False
                            )
                        ]
                    except BUIserverException:
                        all_clients = []
                    grants[serv] = [
                        x
                        for x in all_clients
                        if mask.is_client_allowed(current_user, x, serv)
                    ]
                else:
                    grants[serv] = "all"
            for (serv, clients) in grants.items():
                if not isinstance(clients, list):
                    if data and serv in data:
                        clients = data[serv].keys()
                    else:
                        clients = [
                            x["name"]
                            for x in bui.client.get_all_clients(
                                agent=serv, last_attempt=False
                            )
                        ]
                for cl in clients:
                    (color, text) = self.gen_colors(cl, serv)
                    feed = {
                        "events": self.gen_events(cl, moments, serv, data),
                        "textColor": text,
                        "color": color,
                        "name": "{} on {}".format(cl, serv),
                    }
                    ret.append(feed)

        return ret

    def gen_colors(self, client=None, agent=None):
        """Generates color for an events feed"""
        cache = self._get_color_session(client, agent)
        if cache:
            return (cache["color"], cache["text"])
        labels = bui.client.get_client_labels(client, agent)
        HTML_COLOR = r"((?P<hex>#(?P<red_hex>[0-9a-f]{1,2})(?P<green_hex>[0-9a-f]{1,2})(?P<blue_hex>[0-9a-f]{1,2}))|(?P<rgb>rgb\s*\(\s*(?P<red>2[0-5]{2}|2[0-4]\d|[0-1]?\d\d?)\s*,\s*(?P<green>2[0-5]{2}|2[0-4]\d|[0-1]?\d\d?)\s*,\s*(?P<blue>2[0-5]{2}|2[0-4]\d|[0-1]?\d\d?)\s*\))|(?P<plain>[\w-]+$))"
        color_found = False
        color = None
        text = None
        for label in labels:
            # We are looking for labels starting with "color:" or "text:"
            if re.search(r"^color:", label, re.IGNORECASE):
                search = re.search(
                    r"^color:\s*{}".format(HTML_COLOR), label, re.IGNORECASE
                )
                # we allow various color forms. For instance:
                # hex: #fa12e6
                # rgb: rgb (123, 42, 9)
                # plain: black
                if search.group("hex"):
                    red = search.group("red_hex")
                    green = search.group("green_hex")
                    blue = search.group("blue_hex")
                    # ensure ensure the hex part is of the form XX
                    red = red + red if len(red) == 1 else red
                    green = green + green if len(green) == 1 else green
                    blue = blue + blue if len(blue) == 1 else blue
                    # Now convert the hex to an int
                    red = int(red, 16)
                    green = int(green, 16)
                    blue = int(blue, 16)
                elif search.group("rgb"):
                    red = int(search.group("red"))
                    green = int(search.group("green"))
                    blue = int(search.group("blue"))
                elif search.group("plain"):
                    # if plain color is provided, we cannot guess the adapted
                    # text color, so we assume white (unless text is specified)
                    red = 0
                    green = 0
                    blue = 0
                    color = search.group("plain")
                else:
                    continue
                color = color or "#{:02X}{:02X}{:02X}".format(red, green, blue)
                color_found = True
            if re.search(r"^text:", label, re.IGNORECASE):
                search = re.search(
                    r"^text:\s*{}".format(HTML_COLOR), label, re.IGNORECASE
                )
                # if we don't find anything, we'll generate a color based on
                # the value of the red, green and blue variables
                text = (
                    search.group("hex") or search.group("rgb") or search.group("plain")
                )
            if color and text:
                break

        if not color_found:

            def rand():
                return random.randint(0, 255)

            red = rand()
            green = rand()
            blue = rand()

        text = text or self._get_text_color(red, green, blue)
        color = color or "#{:02X}{:02X}{:02X}".format(red, green, blue)
        self._set_color_session(color, text, client, agent)
        return (color, text)

    def _get_text_color(self, red=0, green=0, blue=0):
        """Generates the text color for a given color"""
        yiq = ((red * 299) + (green * 587) + (blue * 114)) / 1000
        return "black" if yiq >= 128 else "white"

    def _get_color_session(self, client, agent=None):
        """Since we can *paginate* the rendering, we need to store the already
        generated colors

        This method allows to retrieve already generated colors if any
        """
        sess = session._get_current_object()
        if "colors" in sess:
            colors = sess["colors"]
            if agent and agent in colors:
                return colors[agent].get(client)
            elif not agent:
                return colors.get(client)
        return None

    def _set_color_session(self, color, text, client, agent=None):
        """Since we can *paginate* the rendering, we need to store the already
        generated colors

        This method allows to store already generated colors in the session
        """
        sess = session._get_current_object()
        dic = {}
        if agent:
            if "colors" in sess and agent in sess["colors"]:
                dic[agent] = sess["colors"][agent]
            else:
                dic[agent] = {}
            dic[agent][client] = {"color": color, "text": text}
        else:
            dic[client] = {"color": color, "text": text}
        if "colors" in sess:
            sess["colors"].update(dic)
        else:
            sess["colors"] = dic

    def gen_events(self, client, moments, server=None, data=None):
        """Creates events for a given client"""
        events = []
        filtered = False
        if data:
            if bui.config["STANDALONE"]:
                events = data.get(client, [None])
            else:
                events = data.get(server, {}).get(client, [None])
        if not events:
            events = bui.client.get_client_filtered(
                client, start=moments["start"], end=moments["end"], agent=server
            )
            filtered = True

        ret = []
        for ev in events:
            if not ev:
                continue
            if data and not filtered:
                # events are sorted by date DESC
                if moments["start"] and ev["date"] < moments["start"]:
                    continue
                if moments["end"] and ev["date"] > moments["end"]:
                    continue
            ev["title"] = "Client: {0}, Backup n°{1:07d}".format(
                client, int(ev["number"])
            )
            if server:
                ev["title"] += ", Server: {0}".format(server)
            ev["name"] = client
            ev["url"] = url_for(
                "view.backup_report",
                name=client,
                server=server,
                backup=int(ev["number"]),
            )
            ret.append(ev)

        return ret
